#This centralizes some of the common elements that are differently named across solvers.
#These are stored as strings here to prevent problems associated with calling
#solver objects for solver packages that aren't available
from copy import deepcopy
__objective_sense_cplex = {'maximize': 'Cplex.objective.sense.maximize',
                           'minimize': 'Cplex.objective.sense.minimize'}
__objective_sense_cplex_java = {'maximize': 'IloObjectiveSense.Maximize',
                                'minimize': 'IloObjectiveSense.Minimize'}

__objective_sense_glpk = {'maximize': True,
                          'minimize': False}
__objective_sense_glpk_java = {'maximize': 'GLPKConstants.GLP_MAX',
                               'minimize': 'GLPKConstants.GLP_MIN'}
__objective_sense_gurobi = {'maximize': -1,
                            'minimize': 1}
objective_senses = {'cplex': __objective_sense_cplex,
                    'cplex_java': __objective_sense_cplex_java,
                    'glpk': __objective_sense_glpk,
                    'glpk_java': __objective_sense_glpk_java,
                    'gurobi': __objective_sense_gurobi}
default_objective_sense = 'maximize'
#Mappings from solver-specific status values to cobra pie status values
## __status_cplex = "{Cplex.solution.status.MIP_infeasible: 'infeasible', " +\
##                  "Cplex.solution.status.MIP_optimal: 'optimal', " +\
##                  "Cplex.solution.status.MIP_optimal_tolerance: 'optimal'," +\
##                  "Cplex.solution.status.MIP_unbounded:  'unbounded', "+\
##                  "Cplex.solution.status.infeasible: 'infeasible', " +\
##                  "Cplex.solution.status.optimal: 'optimal',  " +\
##                  "Cplex.solution.status.optimal_tolerance: 'optimal', " +\
##                  "Cplex.solution.status.unbounded: 'unbounded', }"
__status_cplex = """{
    'MIP_infeasible': 'infeasible',
    'integer optimal solution': 'optimal',
    'MIP_optimal': 'optimal',
    'MIP_optimal_tolerance': 'optimal',
    'MIP_unbounded':  'unbounded',
    'infeasible': 'infeasible',
    'optimal': 'optimal',
    'optimal_tolerance': 'optimal',
    'unbounded': 'unbounded',
    'integer optimal, tolerance': 'optimal',
    'time limit exceeded': 'time_limit'
}"""

__status_glpk = """{
    'opt': 'optimal',
    'nofeas': 'infeasible',
    'unbnd': 'unbounded'
}"""
__status_glpk_java = "{GLPKConstants.GLP_OPT: 'optimal', GLPKConstants.GLP_FEAS: 'feasible', GLPKConstants.GLP_INFEAS: 'infeasible', GLPKConstants.GLP_NOFEAS: 'infeasible', GLPKConstants.GLP_UNBND: 'unbounded', GLPKConstants.GLP_UNDEF: 'undefined'}"
__status_gurobi = "{GRB.OPTIMAL: 'optimal', GRB.INFEASIBLE: 'infeasible', GRB.UNBOUNDED: 'unbounded', GRB.TIME_LIMIT: 'time_limit'}"

status_dict = {'cplex': __status_cplex,
               'glpk': __status_glpk,
               'glpk_java': __status_glpk_java,
               'gurobi': __status_gurobi}

#Mappings from solver-specific variable kinds to cobra pie
__kind_cplex = "{'continuous': Cplex.variables.type.continuous, 'integer': Cplex.variables.type.integer}"
__kind_cplex_java = "{'continuous':  IloNumVarType.Float, 'integer': IloNumVarType.Int}"
__kind_glpk = "{'continuous': float, 'integer': int}"
__kind_glpk_java = "{'binary': GLPKConstants.GLP_BV, 'continuous': GLPKConstants.GLP_CV, 'integer': GLPKConstants.GLP_IV}"
__kind_gurobi = "{'continuous': GRB.CONTINUOUS, 'integer': GRB.INTEGER}"

variable_kind_dict = {'cplex': __kind_cplex,
                      'cplex_java': __kind_cplex_java,
                      'glpk': __kind_glpk,
                      'glpk_java': __kind_glpk_java, 
                      'gurobi': __kind_gurobi}

#Mappings from solver-specific constraint senses to cobra pie
sense_dict = {'cplex': "{'E': 'E', 'L': 'L', 'G': 'G'}",
              'glpk': "{'E': 'E', 'L': 'L', 'G': 'G'}",
              'gurobi': "{'E': GRB.EQUAL, 'L': GRB.LESS_EQUAL, 'G': GRB.GREATER_EQUAL}"}


#Mappings from cobra pie parameters names to solver specific parameter names
__mappings_cplex = {'lp_method': 'lpmethod',
                    'lp_parallel': 'threads',
                    'threads': 'threads',
                    'objective_sense': 'objective_sense',
                    'time_limit': 'timelimit',
                    'iteration_limit': 'simplex.limits.iterations',
                    'tolerance_barrier': 'barrier.convergetol',
                    'tolerance_feasibility': 'simplex.tolerances.feasibility',
                    'tolerance_markowitz': 'simplex.tolerances.markowitz',
                    'tolerance_optimality': 'simplex.tolerances.optimality',
                    'MIP_gap_abs': 'mip.tolerances.absmipgap',
                    'MIP_gap': 'mip.tolerances.mipgap'}
__mappings_cplex_java = {'lp_method': 'RootAlg',
                         'lp_parallel': 'ParallelMode',
                         'objective_sense': 'objective_sense',
                         'time_limit': 'TiLim',
                         'tolerance_barrier': 'BarEpComp',
                         'tolerance_feasibility': 'EpRHS',
                         'tolerance_markowitz': 'EpMrk',
                         'tolerance_optimality': 'EpOpt'}
__mappings_glpk = {}
__mappings_glpk_java = {'objective_sense': 'objective_sense',
                        'lp_method': 'meth',
                        'output_verbosity': 'msg_lev',
                        'tolerance_dual': 'tol_dj',
                        'tolerance_integer': 'tol_int',
                        'tolerance_optimality': 'tol_bnd'
                        }
__mappings_gurobi = {'log_file': 'LogFile',
                     'lp_method': 'Method',
                     'threads': 'Threads',
                     'objective_sense': 'ModelSense',
                     'output_verbosity': 'OutputFlag',
                     'quadratic_precision': 'Quad',
                     'time_limit': 'TimeLimit',
                     'tolerance_feasibility': 'FeasibilityTol',
                     'tolerance_markowitz': 'MarkowitzTol',
                     'tolerance_optimality': 'OptimalityTol',
                     'iteration_limit': 'IterationLimit',
                     'MIP_gap_abs': 'MIPGapAbs',
                     'MIP_gap': 'MIPGap'}
parameter_mappings = {'cplex': __mappings_cplex,
                      'cplex_java': __mappings_cplex_java,
                      'glpk': __mappings_glpk,
                      'glpk_java': __mappings_glpk_java,
                      'gurobi': __mappings_gurobi}


#Default solver parameters
__common_defaults = {'objective_sense': 'maximize',
                      'tolerance_optimality': 1e-6, 'tolerance_feasibility': 1e-6,
                      'tolerance_integer': 1e-9, 
                      'quadratic_component': None}
                      


__parameters_cplex = deepcopy(__common_defaults)
__parameters_cplex.update({'lp_method': 1,
                           'lp_parallel': 0,
                           'tolerance_barrier': 1.e-8})
__parameters_glpk = deepcopy(__common_defaults)
__parameters_glpk.update({'lp_method': 1})
__parameters_glpk_java = deepcopy(__common_defaults)
__parameters_glpk_java.update({'lp_method': 1,
                               'output_verbosity': 0,
                               'tolerance_dual': 1e-8})
__parameters_gurobi = deepcopy(__common_defaults)
__parameters_gurobi.update({'output_verbosity': 0,
                          'lp_method': 0,
                          'log_file': '',
                          'tolerance_barrier': 1e-8})


parameter_defaults = {'cplex': __parameters_cplex,
                      'glpk': __parameters_glpk,
                      'glpk_java': __parameters_glpk_java,
                      'gurobi': __parameters_gurobi}



#cobra/sbml.py: Tools for reading / writing SBML now contained in
#this module
#System modules
from .. import Model, Reaction, Metabolite
from os.path import isfile
from os import name as __name
from warnings import warn
import re
from math import isnan, isinf

from six import iteritems

#
if __name == 'java':
    from org.sbml.jsbml import SBMLDocument, SpeciesReference, KineticLaw, Parameter
    from org.sbml.jsbml import SBMLReader, SBMLWriter
    __tmp_reader = SBMLReader()
    __tmp_writer = SBMLWriter()
    readSBML = __tmp_reader.readSBMLFromFile
    writeSBML = __tmp_writer.writeSBMLToFile
    from org.sbml.jsbml.Unit import Kind as __Kind
    UNIT_KIND_MOLE = __Kind.MOLE
    UNIT_KIND_GRAM = __Kind.GRAM
    UNIT_KIND_SECOND = __Kind.SECOND
    UNIT_KIND_DIMENSIONLESS = __Kind.DIMENSIONLESS
else:
    from libsbml import SBMLDocument, SpeciesReference, KineticLaw, Parameter
    from libsbml import readSBML, writeSBML
    from libsbml import UNIT_KIND_MOLE, UNIT_KIND_GRAM, UNIT_KIND_SECOND, UNIT_KIND_DIMENSIONLESS
def parse_legacy_id(the_id, the_compartment=None, the_type='metabolite',
                    use_hyphens=False):
    """Deals with a bunch of problems due to bigg.ucsd.edu not following SBML standards

    the_id: String.

    the_compartment: String.

    the_type: String.  Currently only 'metabolite' is supported

    use_hyphens:   Boolean.  If True, double underscores (__) in an SBML ID will be converted to hyphens
    
    """
    if use_hyphens:
        the_id = the_id.replace('__','-')
    if the_type == 'metabolite':
        if the_id.split('_')[-1] == the_compartment:
            #Reformat Ids to match convention in Palsson Lab.
            the_id = the_id[:-len(the_compartment)-1]
        the_id += '[%s]'%the_compartment
    return the_id
def create_cobra_model_from_sbml_file(sbml_filename, old_sbml=False, legacy_metabolite=False,
                                      print_time=False, use_hyphens=False):
    """convert an SBML XML file into a cobra.Model object.  Supports
    SBML Level 2 Versions 1 and 4.  The function will detect if the SBML fbc package is used in the file
    and run the converter if the fbc package is used.

    sbml_filename: String.

    old_sbml:  Boolean. Set to True if the XML file has metabolite
    formula appended to metabolite names.  This was a poorly designed
    artifact that persists in some models.

    legacy_metabolite: Boolean.  If True then assume that the metabolite id has
    the compartment id appended after an underscore (e.g. _c for cytosol).  This
    has not been implemented but will be soon.

    print_time: deprecated

    use_hyphens:   Boolean.  If True, double underscores (__) in an SBML ID will be converted to hyphens

    """
    __default_lower_bound = -1000
    __default_upper_bound = 1000
    __default_objective_coefficient = 0
     # Ensure that the file exists
    if not isfile(sbml_filename):
        raise IOError('Your SBML file is not found: %s'%sbml_filename)
    #Expressions to change SBML Ids to Palsson Lab Ids
    metabolite_re = re.compile('^M_')
    reaction_re = re.compile('^R_')
    compartment_re = re.compile('^C_')
    if print_time:
        warn("print_time is deprecated")
    model_doc = readSBML(sbml_filename)
    if (model_doc.getPlugin("fbc") != None):
        from libsbml import ConversionProperties, LIBSBML_OPERATION_SUCCESS
        conversion_properties = ConversionProperties()
        conversion_properties.addOption("convert fbc to cobra", True, "Convert FBC model to Cobra model")
        result = model_doc.convert(conversion_properties)
        if result != LIBSBML_OPERATION_SUCCESS:
            raise(Exception("Conversion of SBML+fbc to COBRA failed"))
    sbml_model = model_doc.getModel()
    sbml_model_id = sbml_model.getId()
    sbml_species = sbml_model.getListOfSpecies()
    sbml_reactions = sbml_model.getListOfReactions()
    sbml_compartments = sbml_model.getListOfCompartments()
    compartment_dict = dict([(compartment_re.split(x.getId())[-1], x.getName())
                             for x in sbml_compartments])
    if legacy_metabolite:
        #Deal with the palsson lab appending the compartment id to the metabolite id
        new_dict = {}
        for the_id, the_name in compartment_dict.items():
            if the_name == '':
                new_dict[the_id[0].lower()] = the_id
            else:
                new_dict[the_id] = the_name
        compartment_dict = new_dict
        legacy_compartment_converter = dict([(v,k)
                                             for k, v in compartment_dict.items()])

    cobra_model = Model(sbml_model_id)
    metabolites = []
    metabolite_dict = {}
    #Convert sbml_metabolites to cobra.Metabolites
    for sbml_metabolite in sbml_species:
        #Skip sbml boundary species
        if sbml_metabolite.getBoundaryCondition():
            continue

        if (old_sbml or legacy_metabolite) and \
               sbml_metabolite.getId().endswith('_b'):
            #Deal with incorrect sbml from bigg.ucsd.edu
            continue
        tmp_metabolite = Metabolite()
        metabolite_id = tmp_metabolite.id = sbml_metabolite.getId()
        tmp_metabolite.compartment = compartment_re.split(sbml_metabolite.getCompartment())[-1]
        if legacy_metabolite:
            if tmp_metabolite.compartment not in compartment_dict:
                tmp_metabolite.compartment = legacy_compartment_converter[tmp_metabolite.compartment]
            tmp_metabolite.id = parse_legacy_id(tmp_metabolite.id, tmp_metabolite.compartment,
                                                use_hyphens=use_hyphens)
        if use_hyphens:
            tmp_metabolite.id = metabolite_re.split(tmp_metabolite.id)[-1].replace('__','-')
        else:
            #Just in case the SBML ids are ill-formed and use -
            tmp_metabolite.id = metabolite_re.split(tmp_metabolite.id)[-1].replace('-','__')
        tmp_metabolite.name = sbml_metabolite.getName()
        tmp_formula = ''
        tmp_metabolite.notes = parse_legacy_sbml_notes(sbml_metabolite.getNotesString())
        if sbml_metabolite.isSetCharge():
            tmp_metabolite.charge = sbml_metabolite.getCharge()
        if "CHARGE" in tmp_metabolite.notes:
            note_charge = tmp_metabolite.notes["CHARGE"][0]
            try:
                note_charge = float(note_charge)
                if note_charge == int(note_charge):
                    note_charge = int(note_charge)
            except:
                warn("charge of %s is not a number (%s)" % (tmp_metabolite.id, str(note_charge)))
            else:
                if tmp_metabolite.charge is None or tmp_metabolite.charge == note_charge:
                    tmp_metabolite.notes.pop("CHARGE")
                else:  # tmp_metabolite.charge != note_charge
                    msg = "different charges specified for %s (%d and %d)"
                    msg = msg % (tmp_metabolite.id, tmp_metabolite.charge, note_charge)
                    warn(msg)

        for the_key in tmp_metabolite.notes.keys():
            if the_key.lower() == 'formula':
                tmp_formula = tmp_metabolite.notes.pop(the_key)[0]
                break
        if tmp_formula == '' and old_sbml:
            tmp_formula = tmp_metabolite.name.split('_')[-1]
            tmp_metabolite.name = tmp_metabolite.name[:-len(tmp_formula)-1]
        tmp_metabolite.formula = tmp_formula
        metabolite_dict.update({metabolite_id: tmp_metabolite})
        metabolites.append(tmp_metabolite)
    cobra_model.add_metabolites(metabolites)

    #Construct the vectors and matrices for holding connectivity and numerical info
    #to feed to the cobra toolbox.
    #Always assume steady state simulations so b is set to 0
    cobra_reaction_list = []
    for sbml_reaction in sbml_reactions:
        if use_hyphens:
            #Change the ids to match conventions used by the Palsson lab.
            reaction = Reaction(reaction_re.split(sbml_reaction.getId())[-1].replace('__','-'))
        else:
            #Just in case the SBML ids are ill-formed and use -
            reaction = Reaction(reaction_re.split(sbml_reaction.getId())[-1].replace('-','__'))
        cobra_reaction_list.append(reaction)
        #reaction.exchange_reaction = 0
        reaction.name = sbml_reaction.getName()
        cobra_metabolites = {}
        #Use the cobra.Metabolite class here
        for sbml_metabolite in sbml_reaction.getListOfReactants():
            tmp_metabolite_id = sbml_metabolite.getSpecies()
            #This deals with boundary metabolites
            if tmp_metabolite_id in metabolite_dict:
                tmp_metabolite = metabolite_dict[tmp_metabolite_id]
                cobra_metabolites[tmp_metabolite] = -sbml_metabolite.getStoichiometry()
        for sbml_metabolite in sbml_reaction.getListOfProducts():
            tmp_metabolite_id = sbml_metabolite.getSpecies()
            #This deals with boundary metabolites
            if tmp_metabolite_id in metabolite_dict:
                tmp_metabolite = metabolite_dict[tmp_metabolite_id]
                # Handle the case where the metabolite was specified both
                # as a reactant and as a product.
                if tmp_metabolite in cobra_metabolites:
                    warn("%s appears as a reactant and product %s" %
                         (tmp_metabolite_id, reaction.id))
                    cobra_metabolites[tmp_metabolite] += sbml_metabolite.getStoichiometry()
                    # if the combined stoichiometry is 0, remove the metabolite
                    if cobra_metabolites[tmp_metabolite] == 0:
                        cobra_metabolites.pop(tmp_metabolite)
                else:
                    cobra_metabolites[tmp_metabolite] = sbml_metabolite.getStoichiometry()
        reaction.add_metabolites(cobra_metabolites)
        #Parse the kinetic law info here.
        parameter_dict = {}
        #If lower and upper bounds are specified in the Kinetic Law then
        #they override the sbml reversible attribute.  If they are not
        #specified then the bounds are determined by getReversible.
        if not sbml_reaction.getKineticLaw():
            
            if sbml_reaction.getReversible():
                parameter_dict['lower_bound'] = __default_lower_bound
                parameter_dict['upper_bound'] = __default_upper_bound
            else:
                #Assume that irreversible reactions only proceed from left to right.
                parameter_dict['lower_bound'] = 0
                parameter_dict['upper_bound'] = __default_upper_bound
                
            parameter_dict['objective_coefficient'] = __default_objective_coefficient
        else:
            for sbml_parameter in sbml_reaction.getKineticLaw().getListOfParameters():
                parameter_dict[sbml_parameter.getId().lower()] = sbml_parameter.getValue()

        if 'lower_bound' in parameter_dict:
            reaction.lower_bound = parameter_dict['lower_bound']
        elif 'lower bound' in parameter_dict:
            reaction.lower_bound = parameter_dict['lower bound']
        elif sbml_reaction.getReversible():
            reaction.lower_bound = __default_lower_bound
        else:
            reaction.lower_bound = 0

        if 'upper_bound' in parameter_dict:
            reaction.upper_bound = parameter_dict['upper_bound']
        elif 'upper bound' in parameter_dict:
            reaction.upper_bound = parameter_dict['upper bound']
        else:
            reaction.upper_bound = __default_upper_bound

        if 'objective_coefficient' in parameter_dict:
            reaction.objective_coefficient = parameter_dict['objective_coefficient']
        elif 'objective coefficient' in parameter_dict:
            reaction.objective_coefficient = parameter_dict['objective coefficient']
        else:
            reaction.objective_coefficient = __default_objective_coefficient

        # ensure values are not set to nan or inf
        if isnan(reaction.lower_bound) or isinf(reaction.lower_bound):
            reaction.lower_bound = __default_lower_bound
        if isnan(reaction.upper_bound) or isinf(reaction.upper_bound):
            reaction.upper_bound = __default_upper_bound

        reaction_note_dict = parse_legacy_sbml_notes(sbml_reaction.getNotesString())
        #Parse the reaction notes.
        #POTENTIAL BUG: DEALING WITH LEGACY 'SBML' THAT IS NOT IN A
        #STANDARD FORMAT
        #TODO: READ IN OTHER NOTES AND GIVE THEM A reaction_ prefix.
        #TODO: Make sure genes get added as objects
        if 'GENE ASSOCIATION' in reaction_note_dict:
            rule = reaction_note_dict['GENE ASSOCIATION'][0]
            try:
                rule.encode('ascii')
            except (UnicodeEncodeError, UnicodeDecodeError):
                warn("gene_reaction_rule '%s' is not ascii compliant" % rule)
            if rule.startswith("&quot;") and rule.endswith("&quot;"):
                rule = rule[6:-6]
            reaction.gene_reaction_rule = rule
            if 'GENE LIST' in reaction_note_dict:
                reaction.systematic_names = reaction_note_dict['GENE LIST'][0]
            elif 'GENES' in reaction_note_dict and \
                     reaction_note_dict['GENES'] != ['']:
                reaction.systematic_names = reaction_note_dict['GENES'][0]
            elif 'LOCUS' in reaction_note_dict:
                gene_id_to_object = dict([(x.id, x) for x in reaction._genes])
                for the_row in reaction_note_dict['LOCUS']:
                    tmp_row_dict = {}
                    the_row = 'LOCUS:' + the_row.lstrip('_').rstrip('#')
                    for the_item in the_row.split('#'):
                        k, v = the_item.split(':')
                        tmp_row_dict[k] = v
                    tmp_locus_id = tmp_row_dict['LOCUS']
                    if 'TRANSCRIPT' in tmp_row_dict:
                        tmp_locus_id = tmp_locus_id + '.' + tmp_row_dict['TRANSCRIPT']
                    
                    if 'ABBREVIATION' in tmp_row_dict:
                        gene_id_to_object[tmp_locus_id].name = tmp_row_dict['ABBREVIATION']

        if 'SUBSYSTEM' in reaction_note_dict:
            reaction.subsystem = reaction_note_dict.pop('SUBSYSTEM')[0]

        reaction.notes = reaction_note_dict


    #Now, add all of the reactions to the model.
    cobra_model.id = sbml_model.getId()
    #Populate the compartment list - This will be done based on cobra.Metabolites
    #in cobra.Reactions in the future.
    cobra_model.compartments = compartment_dict

    cobra_model.add_reactions(cobra_reaction_list)
        #cobra_model.update_rules()
    return cobra_model


def parse_legacy_sbml_notes(note_string, note_delimiter = ':'):
    """Deal with legacy SBML format issues arising from the
    COBRA Toolbox for MATLAB and BiGG.ucsd.edu developers.

    """
    note_dict = {}
    start_tag = '<p>'
    end_tag = '</p>'
    if '<html:p>' in note_string:
        start_tag = '<html:p>'
        end_tag = '</html:p>'
    while start_tag in note_string and end_tag in note_string:
        note_start = note_string.index(start_tag)
        note_end = note_string.index(end_tag)
        the_note = note_string[(note_start + len(start_tag)):note_end].lstrip(' ').rstrip(' ')
        if note_delimiter in the_note:
            note_delimiter_index = the_note.index(note_delimiter)
            note_field = the_note[:note_delimiter_index].lstrip(' ').rstrip(' ').replace('_',' ').upper()
            note_value = the_note[(note_delimiter_index+1):].lstrip(' ').rstrip(' ')
            if note_field in note_dict:
                note_dict[note_field].append(note_value)
            else:
                note_dict[note_field] = [note_value]
        note_string = note_string[(note_end+len(end_tag)): ]

    if 'CHARGE' in note_dict and note_dict['CHARGE'][0].lower() in ['none', 'na', 'nan']:
        note_dict.pop('CHARGE') #Remove non-numeric charges
        
    return(note_dict)

def write_cobra_model_to_sbml_file(cobra_model, sbml_filename,
                                   sbml_level=2, sbml_version=1,
                                   print_time=False,
                                   use_fbc_package=True):
    """Write a cobra.Model object to an SBML XML file.

    cobra_model:  :class:`~cobra.core.Model.Model` object

    sbml_filename:  The file to write the SBML XML to.

    sbml_level:  2 is the only level supported at the moment.

    sbml_version: 1 is the only version supported at the moment.

    use_fbc_package: Boolean.
        Convert the model to the FBC package format to improve portability.
        http://sbml.org/Documents/Specifications/SBML_Level_3/Packages/Flux_Balance_Constraints_(flux)


    TODO: Update the NOTES to match the SBML standard and provide support for
    Level 2 Version 4

    """

    sbml_doc = get_libsbml_document(cobra_model, 
                                   sbml_level=sbml_level, sbml_version=sbml_version,
                                   print_time=print_time,
                                   use_fbc_package=use_fbc_package)

    writeSBML(sbml_doc, sbml_filename)

def get_libsbml_document(cobra_model,
                                   sbml_level=2, sbml_version=1,
                                   print_time=False,
                                   use_fbc_package=True):

    """ Return a libsbml document object for writing to a file. This function
    is used by write_cobra_model_to_sbml_file(). """

    note_start_tag, note_end_tag = '<p>', '</p>'
    if sbml_level > 2 or (sbml_level == 2 and sbml_version == 4):
        note_start_tag, note_end_tag = '<html:p>', '</html:p>'
        
    
    sbml_doc = SBMLDocument(sbml_level, sbml_version)
    sbml_model = sbml_doc.createModel(cobra_model.id.split('.')[0])
    #Note need to set units
    reaction_units = 'mmol_per_gDW_per_hr'
    model_units = sbml_model.createUnitDefinition()
    model_units.setId(reaction_units)
    sbml_unit = model_units.createUnit()
    sbml_unit.setKind(UNIT_KIND_MOLE)
    sbml_unit.setScale(-3)
    sbml_unit = model_units.createUnit()
    sbml_unit.setKind(UNIT_KIND_GRAM)
    sbml_unit.setExponent(-1)
    sbml_unit = model_units.createUnit()
    sbml_unit.setKind(UNIT_KIND_SECOND)
    sbml_unit.setMultiplier(1.0/60/60)
    sbml_unit.setExponent(-1)

    #Add in the common compartment abbreviations.  If there are additional compartments
    #they also need to be added.
    if not cobra_model.compartments:
        cobra_model.compartments = {'c': 'cytosol',
                                    'p': 'periplasm',
                                    'e': 'extracellular'}    
    for the_key in cobra_model.compartments.keys():
        sbml_comp = sbml_model.createCompartment()
        sbml_comp.setId(the_key)
        sbml_comp.setName(cobra_model.compartments[the_key])
        sbml_comp.setSize(1) #Just to get rid of warnings

    if print_time:
        warn("print_time is deprecated")
    #Use this dict to allow for fast look up of species id
    #for references created in the reaction section.
    metabolite_dict = {}

    for cobra_metabolite in cobra_model.metabolites:
        metabolite_dict[cobra_metabolite.id] =  add_sbml_species(sbml_model,
                                                                 cobra_metabolite,
                                                                 note_start_tag=note_start_tag,
                                                                 note_end_tag=note_end_tag)

    for the_reaction in cobra_model.reactions:
        #This is probably the culprit.  Including cobra.Reaction
        #objects explicitly in cobra.Model will speed this up.
        sbml_reaction = sbml_model.createReaction()
        #Need to remove - for proper SBML.  Replace with __
        the_reaction_id = 'R_' + the_reaction.id.replace('-','__' )
        sbml_reaction.setId(the_reaction_id)
        # The reason we are not using the Reaction.reversibility property
        # is because the SBML definition of reversibility does not quite
        # match with the cobra definition. In cobra, reversibility implies
        # that both positive and negative flux values are feasible. However,
        # SBML requires negative-flux-only reactions to still be classified
        # as reversible. To quote from the SBML Level 3 Version 1 Spec:
        # > However, labeling a reaction as irreversible is interpreted as
        # > an assertion that the rate expression will not have negative
        # > values during a simulation.
        # (Page 60 lines 44-45)
        sbml_reaction.setReversible(the_reaction.lower_bound < 0)
        if the_reaction.name:
            sbml_reaction.setName(the_reaction.name)
        else:
            sbml_reaction.setName(the_reaction.id)
        #Add in the reactant/product references
        for the_metabolite, the_coefficient in the_reaction._metabolites.items():
            sbml_stoichiometry = the_coefficient
            metabolite_id = str(metabolite_dict[the_metabolite.id])
            #Each SpeciesReference must have a unique id
            if sbml_stoichiometry < 0:
                species_reference = sbml_reaction.createReactant()
            else:
                species_reference = sbml_reaction.createProduct()
            species_reference.setId(metabolite_id + '_' + the_reaction_id)
            species_reference.setSpecies(metabolite_id)
            species_reference.setStoichiometry(abs(sbml_stoichiometry))
        #Deal with the case where the reaction is a boundary reaction
        if len(the_reaction._metabolites) == 1:
            the_metabolite, the_coefficient = list(the_reaction._metabolites.items())[0]
            metabolite_id = add_sbml_species(sbml_model, the_metabolite,
                                             note_start_tag=note_start_tag,
                                             note_end_tag=note_end_tag,
                                             boundary_metabolite=True)
            sbml_stoichiometry = -the_coefficient
            #Each SpeciesReference must have a unique id
            if sbml_stoichiometry < 0:
                species_reference = sbml_reaction.createReactant()
            else:
                species_reference = sbml_reaction.createProduct()
            species_reference.setId(metabolite_id + '_' + the_reaction_id)
            species_reference.setSpecies(metabolite_id)
            species_reference.setStoichiometry(abs(sbml_stoichiometry))
            
        #Add in the kineticLaw
        sbml_law = KineticLaw(sbml_level, sbml_version)
        if hasattr(sbml_law, 'setId'):
            sbml_law.setId('FLUX_VALUE')
        sbml_law.setFormula('FLUX_VALUE')
        reaction_parameter_dict = {'LOWER_BOUND': [the_reaction.lower_bound, reaction_units],
                                   'UPPER_BOUND': [the_reaction.upper_bound, reaction_units],
                                   'FLUX_VALUE': [0, reaction_units],
                                   'OBJECTIVE_COEFFICIENT': [the_reaction.objective_coefficient,
                                                             'dimensionless']}
        for k, v in reaction_parameter_dict.items():
            sbml_parameter = Parameter(sbml_level, sbml_version)
            sbml_parameter.setId(k)
            if hasattr(v, '__iter__'):
                sbml_parameter.setValue(v[0])
                sbml_parameter.setUnits(v[1])
            else:
                sbml_parameter.setValue(v)
            sbml_law.addParameter(sbml_parameter)
        sbml_reaction.setKineticLaw(sbml_law)

        #Checks if GPR and Subsystem annotations are present in the notes section and if they are the same as those in
        #the reaction's gene_reaction_rule/ subsystem attribute
        #If they are not identical, they are set to be identical
        note_dict = the_reaction.notes.copy()
        if the_reaction.gene_reaction_rule:
            note_dict['GENE_ASSOCIATION'] = [str(the_reaction.gene_reaction_rule)]
        if the_reaction.subsystem:
            note_dict['SUBSYSTEM'] = [str(the_reaction.subsystem)]

        #In a cobrapy model the notes section is stored as a dictionary. The following section turns the key-value-pairs
        #of the dictionary into a string and replaces recurring symbols so that the string has the required syntax for
        #an SBML doc.
        note_str = str(list(iteritems(note_dict)))
        note_start_tag, note_end_tag, note_delimiter = '<p>', '</p>', ':'
        note_str = note_str.replace('(\'',note_start_tag)
        note_str = note_str.replace('\']),',note_end_tag)
        note_str = note_str.replace('\',',note_delimiter)
        note_str = note_str.replace('\']','')
        note_str = note_str.replace('[\'','')
        note_str = note_str.replace('[','<html xmlns="http://www.w3.org/1999/xhtml">')
        note_str = note_str.replace(')]',note_end_tag+'</html>')
        sbml_reaction.setNotes(note_str)

    if use_fbc_package:
        try:
            from libsbml import ConversionProperties, LIBSBML_OPERATION_SUCCESS
            conversion_properties = ConversionProperties()
            conversion_properties.addOption("convert cobra", True, "Convert Cobra model")
            result = sbml_doc.convert(conversion_properties)
            if result != LIBSBML_OPERATION_SUCCESS:
                raise Exception("Conversion of COBRA to SBML+fbc failed")
        except Exception as e:
            error_string = 'Error saving as SBML+fbc. %s'
            try:
                #Check whether the FbcExtension is there
                from libsbml import FbcExtension
                error_string = error_string%e
            except ImportError:
                error_string = error_string%'FbcExtension not available in libsbml. ' +\
                               'If use_fbc_package == True then libsbml must be compiled with ' +\
                               'the fbc extension. '
                from libsbml import getLibSBMLDottedVersion
                _sbml_version = getLibSBMLDottedVersion()
                _major, _minor, _patch = map(int, _sbml_version.split('.'))
                if _major < 5 or (_major == 5 and _minor < 8):
                    error_string += "You've got libsbml %s installed.   You need 5.8.0 or later with the fbc package"

            raise(Exception(error_string))
    return sbml_doc

def add_sbml_species(sbml_model, cobra_metabolite, note_start_tag,
                     note_end_tag, boundary_metabolite=False):
    """A helper function for adding cobra metabolites to an sbml model.

    sbml_model: sbml_model object

    cobra_metabolite: a cobra.Metabolite object

    note_start_tag: the start tag for parsing cobra notes. this will eventually
    be supplanted when COBRA is worked into sbml.

    note_end_tag: the end tag for parsing cobra notes. this will eventually
    be supplanted when COBRA is worked into sbml.

    """
    sbml_species = sbml_model.createSpecies()
    the_id = 'M_' + cobra_metabolite.id.replace('-', '__')
    #Deal with legacy naming issues
    the_compartment = cobra_metabolite.compartment
    if the_id.endswith('[%s]'%the_compartment):
        the_id = the_id[:-len('[%s]'%the_compartment)]
    elif not the_id.endswith('_%s'%the_compartment):
        the_id += '_%s'%the_compartment
    if boundary_metabolite:
        the_id += '_boundary'
    sbml_species.setId(the_id)
    metabolite_id = the_id
    if boundary_metabolite:
        sbml_species.setBoundaryCondition(True)
    if cobra_metabolite.name:
        sbml_species.setName(cobra_metabolite.name)
    else:
        sbml_species.setName(cobra_metabolite.id)
    if the_compartment is not None:
        try:
            sbml_species.setCompartment(the_compartment)
        except:
            warn('metabolite failed: ' + the_id)
            return cobra_metabolite
    if cobra_metabolite.charge is not None:
        sbml_species.setCharge(cobra_metabolite.charge)
    #Deal with cases where the formula in the model is not an object but as a string
    if cobra_metabolite.formula or cobra_metabolite.notes:
        tmp_note =  '<html xmlns="http://www.w3.org/1999/xhtml">'
        if hasattr(cobra_metabolite.formula, 'id'):
            tmp_note += '%sFORMULA: %s%s'%(note_start_tag,
                                              cobra_metabolite.formula.id,
                                              note_end_tag)
        else:
            tmp_note += '%sFORMULA: %s%s'%(note_start_tag,
                                  cobra_metabolite.formula,
                                  note_end_tag)
        if hasattr(cobra_metabolite.notes, 'items'):
            for the_id_type, the_id in cobra_metabolite.notes.items():
                if the_id_type.lower() == 'charge':
                    continue #Use of notes['CHARGE'] has been deprecated in favor of metabolite.charge
                if not isinstance(the_id_type, str):
                    the_id_type = str(the_id_type)
                if hasattr(the_id, '__iter__') and len(the_id) == 1:
                    the_id = the_id[0]
                if not isinstance(the_id, str):
                    the_id = str(the_id)
                tmp_note += '%s%s: %s%s'%(note_start_tag,
                                             the_id_type,
                                             the_id, note_end_tag)
        sbml_species.setNotes(tmp_note + '</html>')
    return metabolite_id


def fix_legacy_id(id, use_hyphens=False, fix_compartments=False):
    id = id.replace('_DASH_', '__')
    id = id.replace('_FSLASH_', '/')
    id = id.replace('_BSLASH_', "\\")
    id = id.replace('_LPAREN_', '(')
    id = id.replace('_LSQBKT_', '[')
    id = id.replace('_RSQBKT_', ']')
    id = id.replace('_RPAREN_', ')')
    id = id.replace('_COMMA_', ',')
    id = id.replace('_PERIOD_', '.')
    id = id.replace('_APOS_', "'")
    id = id.replace('&amp;', '&')
    id = id.replace('&lt;', '<')
    id = id.replace('&gt;', '>')
    id = id.replace('&quot;', '"')
    if use_hyphens:
        id = id.replace('__', '-')
    else:
        id = id.replace("-", "__")
    if fix_compartments:
        if len(id) > 2:
            if (id[-3] == "(" and id[-1] == ")") or \
               (id[-3] == "[" and id[-1] == "]"):
                id = id[:-3] + "_" + id[-2]
    return id


def read_legacy_sbml(filename, use_hyphens=False):
    """read in an sbml file and fix the sbml id's"""
    model = create_cobra_model_from_sbml_file(filename, old_sbml=True)
    for metabolite in model.metabolites:
        metabolite.id = fix_legacy_id(metabolite.id)
    model.metabolites._generate_index()
    for reaction in model.reactions:
        reaction.id = fix_legacy_id(reaction.id)
        if reaction.id.startswith("EX_") and reaction.id.endswith("(e)"):
            reaction.id = reaction.id[:-3] + "_e"
    model.reactions._generate_index()
    # remove boundary metabolites (end in _b and only present in exchanges)
    for metabolite in list(model.metabolites):
        if not metabolite.id.endswith("_b"):
            continue
        if len(metabolite._reaction) == 1:
            if list(metabolite._reaction)[0].id.startswith("EX_"):
                metabolite.remove_from_model()
    model.metabolites._generate_index()
    return model



from os import name as __name
from sys import modules as __modules
from warnings import warn
if __name == 'java':
    warn("%s is not yet supported on jython"%__modules[__name__])

else:
    from reporter_metabolites import *
del __name, __modules



from unittest import TestCase, TestLoader, TextTestRunner

import sys

if __name__ == "__main__":
    sys.path.insert(0, "../..")
    from cobra.test import create_test_model, data_directory
    from cobra.core import Metabolite, Model, Reaction
    from cobra.manipulation import *
    sys.path.pop(0)
else:
    from . import create_test_model, data_directory
    from ..core import Metabolite, Model, Reaction
    from ..manipulation import *


class TestManipulation(TestCase):
    """Test functions in cobra.manipulation"""

    def test_canonical_form(self):
        model = create_test_model("textbook")
        # add G constraint to test
        g_constr = Metabolite("SUCCt2_2__test_G_constraint")
        g_constr._constraint_sense = "G"
        g_constr._bound = 5.0
        model.reactions.get_by_id("SUCCt2_2").add_metabolites({g_constr: 1})
        self.assertAlmostEqual(model.optimize("maximize").f, 0.855, places=3)
        # convert to canonical form
        model = canonical_form(model)
        self.assertAlmostEqual(model.optimize("maximize").f, 0.855, places=3)

    def test_canonical_form_minimize(self):
        model = create_test_model("textbook")
        # make a minimization problem
        model.reactions.get_by_id("Biomass_Ecoli_core").lower_bound = 0.5
        for reaction in model.reactions:
            reaction.objective_coefficient = reaction.id == "GAPD"
        self.assertAlmostEqual(model.optimize("minimize").f, 6.27, places=3)
        # convert to canonical form. Convert minimize to maximize
        model = canonical_form(model, objective_sense="minimize")
        self.assertAlmostEqual(model.optimize("maximize").f, -6.27, places=3)
        # lower bounds should now be <= constraints
        self.assertEqual(
            model.reactions.get_by_id("Biomass_Ecoli_core").lower_bound, 0.0)

    def test_modify_reversible(self):
        model1 = create_test_model("textbook")
        model1.optimize()
        model2 = create_test_model("textbook")
        convert_to_irreversible(model2)
        model2.optimize()
        self.assertAlmostEqual(model1.solution.f, model2.solution.f, places=3)
        revert_to_reversible(model2)
        model2.optimize()
        self.assertAlmostEqual(model1.solution.f, model2.solution.f, places=3)

        # Ensure revert_to_reversible is robust to solutions generated both
        # before and after reversibility conversion, or not solved at all.
        model3 = create_test_model("textbook")
        model3.optimize()
        convert_to_irreversible(model3)
        revert_to_reversible(model3)
        self.assertAlmostEqual(model1.solution.f, model3.solution.f, places=3)

        # test reaction where both bounds are negative
        model4 = create_test_model("textbook")
        glc = model4.reactions.get_by_id("EX_glc__D_e")
        glc.upper_bound = -1
        convert_to_irreversible(model4)
        model4.optimize()
        self.assertAlmostEqual(model1.solution.f, model4.solution.f, places=3)
        glc_rev = model4.reactions.get_by_id(glc.notes["reflection"])
        self.assertEqual(glc_rev.lower_bound, 1)
        self.assertEqual(glc.upper_bound, 0)
        revert_to_reversible(model4)
        self.assertEqual(glc.upper_bound, -1)

    def test_escape_ids(self):
        model = create_test_model('textbook')
        model.reactions.PGI.gene_reaction_rule = "a.b or c"
        self.assertIn("a.b", model.genes)
        escape_ID(model)
        self.assertNotIn("a.b", model.genes)

    def test_gene_knockout_computation(self):
        cobra_model = create_test_model()

        def find_gene_knockout_reactions_fast(cobra_model, gene_list):
            compiled_rules = get_compiled_gene_reaction_rules(
                cobra_model)
            return find_gene_knockout_reactions(
                cobra_model, gene_list,
                compiled_gene_reaction_rules=compiled_rules)

        def get_removed(m):
            return {x.id for x in m._trimmed_reactions}

        def test_computation(m, gene_ids, expected_reaction_ids):
            genes = [m.genes.get_by_id(i) for i in gene_ids]
            expected_reactions = {m.reactions.get_by_id(i)
                                  for i in expected_reaction_ids}
            removed1 = set(find_gene_knockout_reactions(m, genes))
            removed2 = set(find_gene_knockout_reactions_fast(m, genes))
            self.assertEqual(removed1, expected_reactions)
            self.assertEqual(removed2, expected_reactions)
            delete_model_genes(m, gene_ids, cumulative_deletions=False)
            self.assertEqual(get_removed(m), expected_reaction_ids)
            undelete_model_genes(m)

        gene_list = ['STM1067', 'STM0227']
        dependent_reactions = {'3HAD121', '3HAD160', '3HAD80', '3HAD140',
                               '3HAD180', '3HAD100', '3HAD181', '3HAD120',
                               '3HAD60', '3HAD141', '3HAD161', 'T2DECAI',
                               '3HAD40'}
        test_computation(cobra_model, gene_list, dependent_reactions)
        test_computation(cobra_model, ['STM4221'], {'PGI'})
        test_computation(cobra_model, ['STM1746.S'], {'4PEPTabcpp'})
        # test cumulative behavior
        delete_model_genes(cobra_model, gene_list[:1])
        delete_model_genes(cobra_model, gene_list[1:],
                           cumulative_deletions=True)
        delete_model_genes(cobra_model, ["STM4221"],
                           cumulative_deletions=True)
        dependent_reactions.add('PGI')
        self.assertEqual(get_removed(cobra_model), dependent_reactions)
        # non-cumulative following cumulative
        delete_model_genes(cobra_model, ["STM4221"],
                           cumulative_deletions=False)
        self.assertEqual(get_removed(cobra_model), {'PGI'})
        # make sure on reset that the bounds are correct
        reset_bound = cobra_model.reactions.get_by_id("T2DECAI").upper_bound
        self.assertEqual(reset_bound, 1000.)
        # test computation when gene name is a subset of another
        test_model = Model()
        test_reaction_1 = Reaction("test1")
        test_reaction_1.gene_reaction_rule = "eggs or (spam and eggspam)"
        test_model.add_reaction(test_reaction_1)
        test_computation(test_model, ["eggs"], set())
        test_computation(test_model, ["eggs", "spam"], {'test1'})
        # test computation with nested boolean expression
        test_reaction_1.gene_reaction_rule = \
            "g1 and g2 and (g3 or g4 or (g5 and g6))"
        test_computation(test_model, ["g3"], set())
        test_computation(test_model, ["g1"], {'test1'})
        test_computation(test_model, ["g5"], set())
        test_computation(test_model, ["g3", "g4", "g5"], {'test1'})
        # test computation when gene names are python expressions
        test_reaction_1.gene_reaction_rule = "g1 and (for or in)"
        test_computation(test_model, ["for", "in"], {'test1'})
        test_computation(test_model, ["for"], set())
        test_reaction_1.gene_reaction_rule = "g1 and g2 and g2.conjugate"
        test_computation(test_model, ["g2"], {"test1"})
        test_computation(test_model, ["g2.conjugate"], {"test1"})
        test_reaction_1.gene_reaction_rule = "g1 and (try:' or 'except:1)"
        test_computation(test_model, ["try:'"], set())
        test_computation(test_model, ["try:'", "'except:1"], {"test1"})

    def test_remove_genes(self):
        m = Model("test")
        m.add_reactions([Reaction("r" + str(i + 1)) for i in range(8)])
        self.assertEqual(len(m.reactions), 8)
        rxns = m.reactions
        rxns.r1.gene_reaction_rule = "(a and b) or (c and a)"
        rxns.r2.gene_reaction_rule = "(a and b and d and e)"
        rxns.r3.gene_reaction_rule = "(a and b) or (b and c)"
        rxns.r4.gene_reaction_rule = "(f and b) or (b and c)"
        rxns.r5.gene_reaction_rule = "x"
        rxns.r6.gene_reaction_rule = "y"
        rxns.r7.gene_reaction_rule = "x or     z"
        rxns.r8.gene_reaction_rule = ""
        self.assertIn("a", m.genes)
        self.assertIn("x", m.genes)
        remove_genes(m, ["a"], remove_reactions=False)
        self.assertNotIn("a", m.genes)
        self.assertIn("x", m.genes)
        self.assertEqual(rxns.r1.gene_reaction_rule, "")
        self.assertEqual(rxns.r2.gene_reaction_rule, "")
        self.assertEqual(rxns.r3.gene_reaction_rule, "b and c")
        self.assertEqual(rxns.r4.gene_reaction_rule, "(f and b) or (b and c)")
        self.assertEqual(rxns.r5.gene_reaction_rule, "x")
        self.assertEqual(rxns.r6.gene_reaction_rule, "y")
        self.assertEqual(rxns.r7.genes, {m.genes.x, m.genes.z})
        self.assertEqual(rxns.r8.gene_reaction_rule, "")
        remove_genes(m, ["x"], remove_reactions=True)
        self.assertEqual(len(m.reactions), 7)
        self.assertNotIn("r5", m.reactions)
        self.assertNotIn("x", m.genes)
        self.assertEqual(rxns.r1.gene_reaction_rule, "")
        self.assertEqual(rxns.r2.gene_reaction_rule, "")
        self.assertEqual(rxns.r3.gene_reaction_rule, "b and c")
        self.assertEqual(rxns.r4.gene_reaction_rule, "(f and b) or (b and c)")
        self.assertEqual(rxns.r6.gene_reaction_rule, "y")
        self.assertEqual(rxns.r7.gene_reaction_rule, "z")
        self.assertEqual(rxns.r7.genes, {m.genes.z})
        self.assertEqual(rxns.r8.gene_reaction_rule, "")

    def test_SBO_annotation(self):
        model = create_test_model("textbook")
        rxns = model.reactions
        rxns.EX_o2_e.annotation.clear()
        fake_DM = Reaction("DM_h_c")
        model.add_reaction(fake_DM)
        fake_DM.add_metabolites({model.metabolites.get_by_id("h_c"): -1})
        # this exchange will be set wrong. The function should not overwrite
        # an existing SBO annotation
        rxns.get_by_id("EX_h_e").annotation["SBO"] = "SBO:0000628"
        add_SBO(model)
        self.assertEqual(rxns.EX_o2_e.annotation["SBO"], "SBO:0000627")
        self.assertEqual(rxns.DM_h_c.annotation["SBO"], "SBO:0000628")
        self.assertEqual(rxns.EX_h_e.annotation["SBO"], "SBO:0000628")

    def test_validate_reaction_bounds(self):
        model = create_test_model("textbook")
        model.reactions[0].lower_bound = float("-inf")
        model.reactions[1].lower_bound = float("nan")
        model.reactions[0].upper_bound = float("inf")
        model.reactions[1].upper_bound = float("nan")
        errors = check_reaction_bounds(model)
        self.assertEqual(len(errors), 4)

    def test_validate_formula_compartment(self):
        model = create_test_model("textbook")
        model.metabolites[1].compartment = "fake"
        model.metabolites[1].formula = "(a*.bcde)"
        errors = check_metabolite_compartment_formula(model)
        self.assertEqual(len(errors), 2)

    def test_validate_mass_balance(self):
        model = create_test_model("textbook")
        self.assertEqual(len(check_mass_balance(model)), 0)
        # if we remove the SBO term which marks the reaction as
        # mass balanced, then the reaction should be detected as
        # no longer mass balanced
        EX_rxn = model.reactions.query("EX")[0]
        EX_rxn.annotation.pop("SBO")
        balance = check_mass_balance(model)
        self.assertEqual(len(balance), 1)
        self.assertIn(EX_rxn, balance)


# make a test suite to run all of the tests
loader = TestLoader()
suite = loader.loadTestsFromModule(sys.modules[__name__])


def test_all():
    TextTestRunner(verbosity=2).run(suite)

if __name__ == "__main__":
    test_all()



from ..core import Reaction, Metabolite
from ..manipulation.modify import convert_to_irreversible
from six import iteritems


def construct_loopless_model(cobra_model):
    """construct a loopless model

    This adds MILP constraints to prevent flux from proceeding in a loop, as
    done in http://dx.doi.org/10.1016/j.bpj.2010.12.3707
    Please see the documentation for an explanation of the algorithm.

    This must be solved with an MILP capable solver.

    """
    # copy the model and make it irreversible
    model = cobra_model.copy()
    convert_to_irreversible(model)
    max_ub = max(model.reactions.list_attr("upper_bound"))
    # a dict for storing S^T
    thermo_stoic = {"thermo_var_" + metabolite.id: {}
                    for metabolite in model.metabolites}
    # Slice operator is so that we don't get newly added metabolites
    original_metabolites = model.metabolites[:]
    for reaction in model.reactions[:]:
        # Boundary reactions are not subjected to these constraints
        if len(reaction._metabolites) == 1:
            continue
        # populate the S^T dict
        bound_id = "thermo_bound_" + reaction.id
        for met, stoic in iteritems(reaction._metabolites):
                thermo_stoic["thermo_var_" + met.id][bound_id] = stoic
        # I * 1000 > v --> I * 1000 - v > 0
        reaction_ind = Reaction(reaction.id + "_indicator")
        reaction_ind.variable_kind = "integer"
        reaction_ind.upper_bound = 1
        reaction_ub = Metabolite(reaction.id + "_ind_ub")
        reaction_ub._constraint_sense = "G"
        reaction.add_metabolites({reaction_ub: -1})
        reaction_ind.add_metabolites({reaction_ub: max_ub})
        # This adds a compensating term for 0 flux reactions, so we get
        # S^T x - (1 - I) * 1001 < -1 which becomes
        # S^T x < 1000 for 0 flux reactions and
        # S^T x < -1 for reactions with nonzero flux.
        reaction_bound = Metabolite(bound_id)
        reaction_bound._constraint_sense = "L"
        reaction_bound._bound = max_ub
        reaction_ind.add_metabolites({reaction_bound: max_ub + 1})
        model.add_reaction(reaction_ind)
    for metabolite in original_metabolites:
        metabolite_var = Reaction("thermo_var_" + metabolite.id)
        metabolite_var.lower_bound = -max_ub
        model.add_reaction(metabolite_var)
        metabolite_var.add_metabolites(
            {model.metabolites.get_by_id(k): v
             for k, v in iteritems(thermo_stoic[metabolite_var.id])})
    return model



from scipy.sparse import dok_matrix

from ..solvers import get_solver_name, solver_dict


def create_euclidian_moma_model(cobra_model, wt_model=None, **solver_args):
    # make the wild type copy if none was supplied
    if wt_model is None:
        wt_model = cobra_model.copy()
    else:
        wt_model = wt_model.copy()
        # ensure single objective
        wt_obj = wt_model.reactions.query(lambda x: x > 0,
                                          "objective_coefficient")
        if len(wt_obj) != 1:
            raise ValueError("wt_model must have exactly 1 objective, %d found"
                             % len(wt_obj))

    obj = cobra_model.reactions.query(lambda x: x > 0, "objective_coefficient")
    if len(obj) == 1:
        objective_id = obj[0].id
    else:
        raise ValueError("model must have exactly 1 objective, %d found" %
                         len(obj))

    wt_model.optimize(**solver_args)
    for reaction in wt_model.reactions:
        # we don't want delete_model_gene to remove the wt reaction as well
        reaction.gene_reaction_rule = ''
        if reaction.objective_coefficient != 0:
            reaction.objective_coefficient = 0
            reaction.upper_bound = reaction.lower_bound = reaction.x
        reaction.id = "MOMA_wt_" + reaction.id
    for metabolite in wt_model.metabolites:
        metabolite.id = "MOMA_wt_" + metabolite.id
    wt_model.repair()

    # make the moma model by combining both
    moma_model = cobra_model.copy()
    for reaction in moma_model.reactions:
        reaction.objective_coefficient = 0
    moma_model.add_reactions(wt_model.reactions)
    return moma_model, objective_id


def create_euclidian_distance_objective(n_moma_reactions):
    """returns a matrix which will minimze the euclidian distance

    This matrix has the structure
    [ I  -I]
    [-I   I]
    where I is the identity matrix the same size as the number of
    reactions in the original model.

    n_moma_reactions: int
        This is the number of reactions in the MOMA model, which should
        be twice the number of reactions in the original model"""
    if n_moma_reactions % 2 != 0:
        raise ValueError("must be even")
    n_reactions = n_moma_reactions // 2
    Q = dok_matrix((n_reactions * 2, n_reactions * 2))
    for i in range(2 * n_reactions):
        Q[i, i] = 1
    for i in range(n_reactions):
        Q[i, n_reactions + i] = -1
        Q[n_reactions + i, i] = -1
    return Q


def create_euclidian_distance_lp(moma_model, solver):
    Q = create_euclidian_distance_objective(len(moma_model.reactions))
    lp = solver.create_problem(moma_model, objective_sense="minimize",
                               quadratic_component=Q)
    return lp


def solve_moma_model(moma_model, objective_id, solver=None, **solver_args):
    solver = solver_dict[solver if solver else get_solver_name(qp=True)]
    lp = create_euclidian_distance_lp(moma_model, solver=solver)
    solver.solve_problem(lp, **solver_args)
    solution = solver.format_solution(lp, moma_model)
    solution.f = 0. if solution.x_dict is None \
        else solution.x_dict[objective_id]
    moma_model.solution = solution
    return solution


def moma(wt_model, mutant_model, solver=None, **solver_args):
    if "norm_type" in solver_args:
        print("only euclidian norm type supported for moma")
        solver_args.pop("norm_type")
    moma_model, objective_id = create_euclidian_moma_model(mutant_model,
                                                           wt_model)
    return solve_moma_model(moma_model, objective_id,
                            solver=solver, **solver_args)


def moma_knockout(moma_model, moma_objective, reaction_indexes, **moma_args):
    """computes result of reaction_knockouts using moma"""
    n = len(moma_model.reactions) // 2
    # knock out the reaction
    for i in reaction_indexes:
        mutant_reaction = moma_model.reactions[i]
        mutant_reaction.lower_bound, mutant_reaction.upper_bound = (0., 0.)
    result = solve_moma_model(moma_model, moma_objective, **moma_args)
    # reset the knockouts
    for i in reaction_indexes:
        mutant_reaction = moma_model.reactions[i]
        wt_reaction = moma_model.reactions[n + i]
        mutant_reaction.lower_bound = wt_reaction.lower_bound
        mutant_reaction.upper_bound = wt_reaction.upper_bound
    return result



from os.path import isfile
import csv
import re
from warnings import warn

import cobra


def export_flux_distribution(model, filepath):
    """Export flux distribution to import into Simpheny.

    Parameters
    ----------
    model : cobra.Model
    filepath: str

    """
    from simphenyMapping import mapping
    outfile = open(filepath, "w")
    outcsv = csv.writer(outfile, delimiter="\t", lineterminator="\n")
    outcsv.writerow(["Reaction Number", "Flux Value",
                     "Lower Bound", "Upper Bound"])
    for reaction_name, reaction_flux in model.solution.x_dict.iteritems():
        reaction = model.reactions.get_by_id(reaction_name)
        try:
            outcsv.writerow([mapping[reaction_name], reaction_flux,
                reaction.lower_bound, reaction.upper_bound])
        except KeyError, e:
            print "Simpheny id number not found for", e
    outfile.close()


def _header_count(filename):
    """count the number of header lines in a file
    The header is defined as over when a line is found which begins
    with a number"""
    file = open(filename, "r")
    for i, line in enumerate(file):
        if line[0].isdigit():
            file.close()
            return i
    file.close()
    return False


def _open_and_skip_header(filename):
    """returns a csv file with the header skipped"""
    count = _header_count(filename)
    if not count:
        raise (IOError, "%s corrupted" % filename)
    file = open(filename, "r")
    for i in range(count):
        file.readline()
    return csv.reader(file, delimiter="\t")


def _find_metabolites_by_base(base, metabolites):
    """search for all metabolites in the list which match the base name.
    For example, "h2o" will identify both "h2o(c)" and "h2o(e)" """
    search = re.compile(base + "\([a-z]\)")
    found = []
    for the_metabolite in metabolites:
        if search.match(the_metabolite.id) is not None:
            found.append(the_metabolite)
    return found


def read_simpheny(baseName, min_lower_bound=-1000, max_upper_bound=1000,
        maximize_info=True):
    r"""Imports files exported from a SimPheny simulation as a cobra model.

    .. warning:: Use with caution. This is a legacy import function, and
        errors have been observed in the converted gene-reaction rules.

    Parameters
    ----------
    baseName : str
        The filepath to the exported SimPheny files without any of the
        extensions. On Windows, it helps if baseName is a raw string
        (i.e. r"Path\\to\\files")
    min_lower_bound, max_upper_bound : float or int, optional
        The bounds on the lower and upper bounds of fluxes in model.
    maximize_info : bool, optional
        An optional boolean keyword argument. If True, then an attempt
        will be made to parse the gpr and metabolite info files, and the
        function will take a little bit longer.

    Returns
    -------
    model : cobra.Model
        the imported simpheny model

    """

    # check to make sure the files can be read
    if not(isfile(baseName + ".met")
        and isfile(baseName + ".rxn")
        and isfile(baseName + ".sto")):
        # try again with modifying the baseName
        baseName = baseName.encode("string-escape")
        if not(isfile(baseName + ".met")
            and isfile(baseName + ".rxn")
            and isfile(baseName + ".sto")):
            raise (IOError, "Input file(s) not found")
    model = cobra.Model("SimPheny import from " + baseName)
    # read in metabolite file
    metfile = _open_and_skip_header(baseName + ".met")
    metabolites = []
    for line in metfile:
        if len(line) == 0:
            break
        metabolite = cobra.Metabolite(id=line[1], name=line[2],
            compartment=line[3])
        if maximize_info:
            compartment_search = re.findall("\([a-z]\)$", metabolite.id)
            if compartment_search != []:
                metabolite.compartment = compartment_search[0][1]
                model.compartments[metabolite.compartment] = line[3]
        metabolites.append(metabolite)
    model.add_metabolites(metabolites)
    # scalefunc will limit the maximum and minumum fluxes
    scalefunc = lambda x: max(min(max_upper_bound, x), min_lower_bound)
    # read in reaction file
    reaction_file = _open_and_skip_header(baseName + ".rxn")
    reactions = []
    for line in reaction_file:
        if len(line) == 0:
            break
        the_reaction = cobra.Reaction()
        the_reaction.id = line[1]
        the_reaction.name = line[2]
        if line[3].lower() == "reversible":
            the_reaction.reversibility = 1
        elif line[3].lower() == "irreversible":
            the_reaction.reversibility = 0
        the_reaction.lower_bound = scalefunc(float(line[4]))
        the_reaction.upper_bound = scalefunc(float(line[5]))
        the_reaction.objective_coefficient = float(line[6])
        reactions.append(the_reaction)
    model.add_reactions(reactions)
    # read in S matrix
    Sfile = _open_and_skip_header(baseName + ".sto")
    S = []
    for i, line in enumerate(Sfile):
        if len(line) == 0:
            break
        the_metabolite = metabolites[i]
        for j, ns in enumerate(line):
            n = float(ns)
            if n != 0:
                model.reactions[j].add_metabolites({the_metabolite: n})
    # attempt to read in more data
    infofilepath = baseName + "_cmpd.txt"
    if maximize_info and isfile(infofilepath):
        infofile = open(infofilepath, "r")
        infofile.readline()  # skip the header
        infocsv = csv.reader(infofile)
        for row in infocsv:
            found = _find_metabolites_by_base(row[0], model.metabolites)
            for found_metabolite in found:
                found_metabolite.formula = row[2]
                found_metabolite.parse_composition()
                found_metabolite.charge = row[4]
                found_metabolite.notes = {}
                found_metabolite.notes["KEGG_id"] = row[8]
                found_metabolite.notes["CAS"] = row[5]
                found_metaboltie.notes["review status"] = row[3]
        infofile.close()
    gpr_filepath = baseName + "_gpr.txt"
    if maximize_info and isfile(gpr_filepath):
        warn("SimPheny export files may have errors in the gpr.")
        # Using this may be risky
        gpr_file = open(gpr_filepath, "r")
        gpr_file.readline()  # skip the header
        gpr_csv = csv.reader(gpr_file)
        for row in gpr_csv:
            the_reaction = model.reactions[model.reactions.index(row[0])]
            the_reaction.gene_reaction_rule = row[5]
            the_reaction.parse_gene_association()
        gpr_file.close()
    # model.update()
    return model





from time import time
from numpy import zeros, matrix, array
from scipy.sparse import lil_matrix, dok_matrix
from cPickle import load, dump
from collections import defaultdict
from warnings import warn
def create_element_matrix(cobra_model, db_cursor=None, me_matrix=False,
                          cas='sympy'):
    """Constructs a matrix of elements x metabolites for the metabolites
    in cobra_model.  If elemental compositions are not available for all
    metabolites then a symbolic matrix is returned with symbols representing
    the unknown compositions.

    cobra_model:  A cobra.Model object.

    db_cursor:  Internal use only.

    me_matrix: False. Internal use only.

    cas: 'sympy' or 'ginac'.  Specifies the computer algebra system to use.
    'ginac' is the more powerful solver however it is accessed through
    swiginac which isn't the easiest thing to install.

    """
    if cas.lower() == 'sympy':
        from sympy import solve, Matrix, Symbol, Add
        from sympy.core.numbers import Zero, Real
    elif cas.lower() == 'ginac':
        #Symbolic.Matrix might be slow to index so we might use
        #the swiginac matrix or numpy.matrix instead
        from Symbolic import Symbol, Matrix, Expr
                
        
    elements = ('c', 'h', 'o', 'n', 'p', 's', 'z') #z is for generics
    element_dict = dict(zip(elements, range(len(elements))))
    element_matrix = dok_matrix((len(elements), len(cobra_model.metabolites)))
    if db_cursor and me_matrix:
        #Used for current incarnation of ME matrix.  
        known_compositions = set()
        #1. Start off by getting known molecular compositions
        db_cursor.execute('Select Id, (c, h, o, n, p, s) from metabolite')
        metabolite_compositions = dict(db_cursor.fetchall())
        for the_metabolite, the_composition in metabolite_compositions.items():
            if the_metabolite in cobra_model.metabolites:
                known_compositions.add(the_metabolite)
                the_column = cobra_model.metabolites.index(the_metabolite)
                the_composition = eval(the_composition)
                element_matrix.update(dict([((i, the_column), the_composition[i])
                                            for i in range(len(the_composition))]))

        #2. Identify the reactions that produce generics and set the stoichiometry
        #to zero to deal with mass imbalances.
        #
        #This isn't a problem any more as the dummy reactions are not added
        #until after balancing.
        generic_production_reactions = dict([(x, cobra_model.reactions.index(x))
                                             for x in cobra_model.reactions\
                                             if x.startswith('generic_rename')])
        #collect all of the generic_X metabolite Ids and then set masses to 1 Z
        db_cursor.execute('SELECT FU_ID_Generic from Generic_FU')
        generic_metabolites = [x[0] for x in db_cursor.fetchall()]

        known_compositions.update(generic_metabolites)
        element_matrix.update(dict([((elements.index('z'),
                                      cobra_model.metabolites.index(x)), 1.)
                                    for x in generic_metabolites]))
        #Remove the generic production reactions
        for column_index in generic_production_reactions.values():
            the_column = cobra_model._S[:, column_index]
            for row_index in the_column.nonzero()[0]:
                the_column[row_index, 0] = 0

        #3. Remove demand reactions.
        #This isn't a problem any more as the demand reactions are not added
        #until after balancing.
        demand_indices = [cobra_model.reactions.index(x)
                          for x in cobra_model.reactions if 'demand' in x]
        for column_index in demand_indices:
            the_column = cobra_model._S[:, column_index]
            for row_index in the_column.nonzero()[0]:
                the_column[row_index, 0] = 0
        #4. Calculate molecular formula for transcripts and peptides.  This isn't
        #necessary, but will probably make solving the problem easier.

        #known_compositions.update(transcripts and peptides)

        #5. For all metabolites not in known compositions.  It will be
        #necessary to add a symbolic value to the matrix for each element
        #excluding z, i.e.  the_metabolite_(c, h, o, n, p, s, z). 
        metabolite_dict = dict(zip(cobra_model.metabolites,
                                   range(len(cobra_model.metabolites))))
        [metabolite_dict.pop(k)
         for k in known_compositions]
        #If there are any metabolites without compositions build a symbolic matrix
        if len(metabolite_dict) > 0:
            #Build the symbolic elemental composition matrix
            if cas.lower() == 'sympy':
                element_matrix = Matrix(element_matrix.todense())
            elif cas.lower() == 'ginac':
                element_matrix = Matrix(element_matrix.todense().tolist())
            #Now populate with symbols for the_metabolites not in known_compositions
            for the_metabolite, metabolite_index in metabolite_dict.items():
                for the_element, element_index in element_dict.items():
                    element_matrix[element_index,
                                   metabolite_index] = Symbol('%s_%s'%(the_metabolite,
                                                                       the_element))
    else:
        print 'Not yet implemented for anything other than ME'

    return({'elements': element_dict,
            'matrix': element_matrix})





def create_balancing_problem(cobra_model, element_vector,
                             system_type='equations',
                             cas='sympy', print_unbalanced=False):
    """Create a symbolic linear algebra problem to solve for unknown metabolite
    compositions in cobra_model.

      element_vector x cobra_model._S.T = 0

      Note: The problems are typically too big to balance for all elements
      at once, thus the vector should only deal with one of the chemical elements.

    cobra_model: A cobra.Model object

    element_vector: A sympy.Matrix with the element counts corresponding to each
    metabolite in cobra_model.reactions.

    return_type: 'equations' or 'matrix'.  Constructing equations is faster, but
    for specific operations a matrix may be desired.  If 'matrix' then the
    unbalanced reactions are not currently returned.

    cas: 'sympy' or 'ginac'.  Specifies the computer algebra system to use.
    'ginac' is the more powerful solver however it is accessed through
    swiginac which isn't the easiest thing to install.

    print_unbalanced: Boolean.  Indicates whether to print the unbalanced reactions.
    
    """
    if cas.lower() == 'sympy':
        from sympy import solve, Matrix, Symbol, Add
        from sympy.core.numbers import Zero, Real
    elif cas.lower() == 'ginac':
        from Symbolic import Symbol, Matrix, Expr
        from swiginac import lsolve, matrix, symbol
        from swiginac import add as Add
    #  If multiple solutions are available, then we'll have to make
    # sure that the sum of each column in element_matrix is >= 1
    #
    #Now deal with multiplying the two matrices.
    unbalanced_dict = {}
    variable_set = set()
    if system_type == 'equations':
        s_matrix_transpose = cobra_model._S.T #Row access is faster for sparse arrays
        #Multiplying symbolic_element_matrix by s_matrix is not an option, due to
        #memory and speed issues.
        the_system = []


        for i in range(s_matrix_transpose.shape[0]):
            the_column = s_matrix_transpose[i, :] 
            #This is faster than multiplying symbolic_element_matrix by
            #the_column by about 3-fold
            the_indices = the_column.nonzero()[1]
            the_factors = [float(the_column[0, x])
                           for x in the_indices]
            if cas.lower() == 'sympy':
                the_variables = [element_vector[0,j]
                                 for j in the_indices]
            elif cas.lower() == 'ginac':
                the_variables = [element_vector[j]
                                 for j in the_indices]

            the_equation = reduce(lambda x,y: x + y,
                                   map(lambda x, y: x*y, the_factors,
                                       the_variables))
            if cas.lower() == 'sympy':
                #this can probably be streamlined for the different CASes
                if isinstance(the_equation, Add):
                    the_system.append(the_equation)
                    the_atoms = list(the_equation.atoms())
                    [variable_set.add(x) for x in the_atoms
                     if isinstance(x, Symbol)]
                elif not isinstance(the_equation, Zero) and \
                         the_equation != 0:
                    unbalanced_dict.update({cobra_model.reactions[i]:
                                            the_equation})
                    if print_unbalanced:
                        print 'Unbalanced Reaction ' +\
                              '%s: Element %s is %s'%(cobra_model.reactions[i],
                                                      the_element,
                                                      repr(the_equation))
            elif cas.lower() == 'ginac':
                if isinstance(the_equation.data, Add):
                    the_system.append(the_equation)
                    [variable_set.add(x) for x in the_variables
                     if isinstance(x, Expr)]
                elif the_equation.eval() != 0:
                    unbalanced_dict.update({cobra_model.reactions[i]:
                                            the_equation})
                    if print_unbalanced:
                        print 'Unbalanced Reaction ' +\
                              '%s: Element %s is %s'%(cobra_model.reactions[i],
                                                      the_element,
                                                      repr(the_equation))

    else:
        print 'Warning this may take 10 Gb RAM and an hour'
        the_system = element_vector * cobra_model._S.todense()

    return({'variables': variable_set,
            'equations': the_system,
            'unbalanced': unbalanced_dict})


def solve_balance_problem(the_equations, the_variables):
    """Solves a systems of linear equations for the variables. Using
    sympy.

    the_equations: A list of sympy.Add equations.

    the_variables: A list of sympy variables (Symbols, Zero, Real, One, ...)
    
    """
    if cas.lower() == 'sympy':
        from sympy import solve
    elif cas.lower() == 'ginac':
        from swiginac import lsolve as solve
    the_solution = solve(the_equations, the_variables)    
    return the_solution


if __name__ == '__main__':
    from sys import argv
    from os.path import lexists
    from time import time
    ## if not len(argv) == 3:
    ##     print 'Need to call the script with the model file name and element'
    ##     print 'e.g. python balance.py cobra_model.pickle c'
    ## model_file = argv[1]
    ## the_element = argv[2]
    cas='ginac'
    me_matrix = True
    the_element = 'c'
    model_file = '/Users/danie/e/builds/cobra_model.pickle'
    system_type = 'equations'
    element_file = '%s.elements.%s'%(model_file, cas)
    problem_file = '%s.%s_problem.%s'%(model_file,
                                       the_element,
                                       cas)
    solution_file = '%s.%s_solution.%s'%(model_file,
                                       the_element,
                                       cas)


    with open(model_file) as in_file:
        cobra_model = load(in_file)

    if lexists(element_file) and not lexists(problem_file):
        #Only load the element file if a problem file does not
        #already exist to reduce memory usage.
        with open(element_file) as in_file:
            the_elements = load(in_file)
    elif not lexists(problem_file):
        #Only build the elements if the problem file doesn't exist
        import pgdb as PgSQL
        start_time = time()
        db_con = PgSQL.connect(database='cobra')
        db_cursor = db_con.cursor()
        the_genus = 'thermotoga'
        db_cursor.execute('Set search_path to ' + the_genus)
        print 'Building element matrix'
        start_time = time()
        the_elements = create_element_matrix(cobra_model, db_cursor=db_cursor,
                                             me_matrix=True, cas=cas)
        if cas.lower() != 'ginac':
            #Can't pickle PySwigObjects
            with open(element_file, 'w') as out_file:
                dump(the_elements, out_file) 
        print 'Element matrix %s created in %f minutes'%(element_file,
                                                       (time()-start_time)/60)
    element_index = the_elements['elements'][the_element]
    element_vector = the_elements['matrix'][element_index, :]
    print 'This problem is symbolic and may take some time to solve'
    if not lexists(problem_file):
        print 'Constructing the problem for element %s. Restart to solve the problem'%the_element
        print 'This process can take 1-100 minutes depending on model size'
        start_time = time()
        
        the_problem = create_balancing_problem(cobra_model,
                                               element_vector,
                                               system_type=system_type,
                                               cas=cas,
                                               print_unbalanced=True)
        if cas.lower() != 'ginac':
            print 'Rerun the script to solve the problem'
            with open(problem_file, 'w') as out_file:
                dump(the_problem, out_file)
        print 'Problem file %s created in %1.2f minutes.'%(problem_file,
                                                            (time() - start_time) / 60)
    
    ## else:
    ##     with open(problem_file) as in_file:
    ##         the_problem = load(in_file)

    ##     print 'Solving the problem %s'%problem_file
    ##     print 'This may take some time'
    ##     start_time = time()
    ##     the_solution = solve_balance_problem(the_problem['equations'],
    ##                                          the_problem['variables'])

    ##     with open(solution_file, 'w') as out_file:
    ##         dump(the_solution, out_file)
    ##     print 'Problem solved. %s created in %1.2f minutes.'%(solution_file,
    ##                                                           (time() - start_time) / 60) 
    
#solve nonsymbolic problems
        ## element_matrix = the_elements['matrix'].tocsr()
        ## elements = the_elements['elements']
        ## reaction_matrix = cobra_model._S.T.tocsr()
        ## print 'Not a symbolic problem'
        ## the_balance = element_matrix * reaction_matrix
        ## the_balance = the_balance.tolil()
        ## for e, i in elements:
        ##     print '%s generated by system: %f'%(e, the_balance[i,:].sum())



from unittest import TestCase, TestLoader, TextTestRunner, skipIf

from warnings import warn
import sys
from os.path import join
from json import load

from six import iteritems

try:
    import numpy
except:
    numpy = None
try:
    import matplotlib
except:
    matplotlib = None

if __name__ == "__main__":
    sys.path.insert(0, "../..")
    from cobra.test import create_test_model, data_directory
    from cobra import Model, Reaction, Metabolite
    from cobra.manipulation import initialize_growth_medium
    from cobra.solvers import solver_dict, get_solver_name
    from cobra.flux_analysis import *
    sys.path.pop(0)
else:
    from . import create_test_model, data_directory
    from .. import Model, Reaction, Metabolite
    from ..manipulation import initialize_growth_medium
    from ..solvers import solver_dict, get_solver_name
    from ..flux_analysis import *


class TestCobraFluxAnalysis(TestCase):
    """Test the simulation functions in cobra.flux_analysis"""

    def setUp(self):
        pass

    def test_pFBA(self):
        model = create_test_model("textbook")
        for solver in solver_dict:
            optimize_minimal_flux(model, solver=solver)
            abs_x = [abs(i) for i in model.solution.x]
            self.assertEqual(model.solution.status, "optimal")
            self.assertAlmostEqual(model.solution.f, 0.8739, places=3)
            self.assertAlmostEqual(sum(abs_x), 518.4221, places=3)

    def test_single_gene_deletion(self):
        cobra_model = create_test_model("textbook")
        # expected knockouts for textbook model
        growth_dict = {"fba": {"b0008": 0.87, "b0114": 0.80, "b0116": 0.78,
                               "b2276": 0.21, "b1779": 0.00},
                       "moma": {"b0008": 0.87, "b0114": 0.71, "b0116": 0.56,
                                "b2276": 0.11, "b1779": 0.00},
                       }

        # MOMA requires cplex or gurobi
        try:
            get_solver_name(qp=True)
        except:
            growth_dict.pop('moma')
        for method, expected in growth_dict.items():
            rates, statuses = single_gene_deletion(cobra_model,
                                                   gene_list=expected.keys(),
                                                   method=method)
            for gene, expected_value in iteritems(expected):
                self.assertEqual(statuses[gene], 'optimal')
                self.assertAlmostEqual(rates[gene], expected_value, places=2)

    def test_single_reaction_deletion(self):
        cobra_model = create_test_model("textbook")
        expected_results = {'FBA': 0.70404, 'FBP': 0.87392, 'CS': 0,
                            'FUM': 0.81430, 'GAPD': 0, 'GLUDy': 0.85139}

        results, status = single_reaction_deletion(
            cobra_model, reaction_list=expected_results.keys())
        self.assertEqual(len(results), 6)
        self.assertEqual(len(status), 6)
        for status_value in status.values():
            self.assertEqual(status_value, "optimal")
        for reaction, value in results.items():
            self.assertAlmostEqual(value, expected_results[reaction], 5)

    def compare_matrices(self, matrix1, matrix2, places=3):
        nrows = len(matrix1)
        ncols = len(matrix1[0])
        self.assertEqual(nrows, len(matrix2))
        self.assertEqual(ncols, len(matrix2[0]))
        for i in range(nrows):
            for j in range(ncols):
                self.assertAlmostEqual(matrix1[i][j], matrix2[i][j],
                                       places=places)

    @skipIf(numpy is None, "double deletions require numpy")
    def test_double_gene_deletion(self):
        cobra_model = create_test_model("textbook")
        genes = ["b0726", "b4025", "b0724", "b0720",
                 "b2935", "b2935", "b1276", "b1241"]
        growth_list = [
            [0.858, 0.857, 0.814, 0.000, 0.858, 0.858, 0.858, 0.858],
            [0.857, 0.863, 0.739, 0.000, 0.863, 0.863, 0.863, 0.863],
            [0.814, 0.739, 0.814, 0.000, 0.814, 0.814, 0.814, 0.814],
            [0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000],
            [0.858, 0.863, 0.814, 0.000, 0.874, 0.874, 0.874, 0.874],
            [0.858, 0.863, 0.814, 0.000, 0.874, 0.874, 0.874, 0.874],
            [0.858, 0.863, 0.814, 0.000, 0.874, 0.874, 0.874, 0.874],
            [0.858, 0.863, 0.814, 0.000, 0.874, 0.874, 0.874, 0.874]]
        solution = double_gene_deletion(cobra_model, gene_list1=genes)
        self.assertEqual(solution["x"], genes)
        self.assertEqual(solution["y"], genes)
        self.compare_matrices(growth_list, solution["data"])
        # test when lists differ slightly
        solution = double_gene_deletion(cobra_model, gene_list1=genes[:-1],
                                        gene_list2=genes,
                                        number_of_processes=1)
        self.assertEqual(solution["x"], genes[:-1])
        self.assertEqual(solution["y"], genes)
        self.compare_matrices(growth_list[:-1], solution["data"])

    @skipIf(numpy is None, "double deletions require numpy")
    def test_double_reaction_deletion(self):
        cobra_model = create_test_model("textbook")
        reactions = ['FBA', 'ATPS4r', 'ENO', 'FRUpts2']
        growth_list = [[0.704, 0.135, 0.000, 0.704],
                       [0.135, 0.374, 0.000, 0.374],
                       [0.000, 0.000, 0.000, 0.000],
                       [0.704, 0.374, 0.000, 0.874]]

        solution = double_reaction_deletion(cobra_model,
                                            reaction_list1=reactions,
                                            number_of_processes=1)
        self.assertEqual(solution["x"], reactions)
        self.assertEqual(solution["y"], reactions)
        self.compare_matrices(growth_list, solution["data"])

    def test_flux_variability(self):
        with open(join(data_directory, "textbook_fva.json"), "r") as infile:
            fva_results = load(infile)

        infeasible_model = create_test_model("textbook")
        infeasible_model.reactions.get_by_id("EX_glc__D_e").lower_bound = 0
        for solver in solver_dict:
            # esolver is really slow
            if solver == "esolver":
                continue
            cobra_model = create_test_model("textbook")
            fva_out = flux_variability_analysis(
                cobra_model, solver=solver,
                reaction_list=cobra_model.reactions[1::3])
            for name, result in iteritems(fva_out):
                for k, v in iteritems(result):
                    self.assertAlmostEqual(fva_results[name][k], v, places=5)

            # ensure that an infeasible model does not run FVA
            self.assertRaises(ValueError, flux_variability_analysis,
                              infeasible_model, solver=solver)

    def test_find_blocked_reactions(self):
        m = create_test_model("textbook")
        result = find_blocked_reactions(m, m.reactions[40:46])
        self.assertEqual(result, ['FRUpts2'])

        result = find_blocked_reactions(m, m.reactions[42:48])
        self.assertEqual(set(result), {'FUMt2_2', 'FRUpts2'})

        result = find_blocked_reactions(m, m.reactions[30:50],
                                        open_exchanges=True)
        self.assertEqual(result, [])

    def test_loopless(self):
        try:
            solver = get_solver_name(mip=True)
        except:
            self.skip("no MILP solver found")
        test_model = Model()
        test_model.add_metabolites(Metabolite("A"))
        test_model.add_metabolites(Metabolite("B"))
        test_model.add_metabolites(Metabolite("C"))
        EX_A = Reaction("EX_A")
        EX_A.add_metabolites({test_model.metabolites.A: 1})
        DM_C = Reaction("DM_C")
        DM_C.add_metabolites({test_model.metabolites.C: -1})
        v1 = Reaction("v1")
        v1.add_metabolites({test_model.metabolites.A: -1,
                            test_model.metabolites.B: 1})
        v2 = Reaction("v2")
        v2.add_metabolites({test_model.metabolites.B: -1,
                            test_model.metabolites.C: 1})
        v3 = Reaction("v3")
        v3.add_metabolites({test_model.metabolites.C: -1,
                            test_model.metabolites.A: 1})
        DM_C.objective_coefficient = 1
        test_model.add_reactions([EX_A, DM_C, v1, v2, v3])
        feasible_sol = construct_loopless_model(test_model).optimize()
        v3.lower_bound = 1
        infeasible_sol = construct_loopless_model(test_model).optimize()
        self.assertEqual(feasible_sol.status, "optimal")
        self.assertEqual(infeasible_sol.status, "infeasible")

    def test_gapfilling(self):
        try:
            solver = get_solver_name(mip=True)
        except:
            self.skip("no MILP solver found")
        m = Model()
        m.add_metabolites(map(Metabolite, ["a", "b", "c"]))
        r = Reaction("EX_A")
        m.add_reaction(r)
        r.add_metabolites({m.metabolites.a: 1})
        r = Reaction("r1")
        m.add_reaction(r)
        r.add_metabolites({m.metabolites.b: -1, m.metabolites.c: 1})
        r = Reaction("DM_C")
        m.add_reaction(r)
        r.add_metabolites({m.metabolites.c: -1})
        r.objective_coefficient = 1

        U = Model()
        r = Reaction("a2b")
        U.add_reaction(r)
        r.build_reaction_from_string("a --> b", verbose=False)
        r = Reaction("a2d")
        U.add_reaction(r)
        r.build_reaction_from_string("a --> d", verbose=False)

        result = gapfilling.growMatch(m, U)[0]
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].id, "a2b")

        # 2 rounds with exchange reactions
        result = gapfilling.growMatch(m, None, ex_rxns=True, iterations=2)
        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0]), 1)
        self.assertEqual(len(result[1]), 1)
        self.assertEqual({i[0].id for i in result},
                         {"SMILEY_EX_b", "SMILEY_EX_c"})

    @skipIf(numpy is None, "phase plane requires numpy")
    def test_phenotype_phase_plane(self):
        model = create_test_model("textbook")
        data = calculate_phenotype_phase_plane(
            model, "EX_glc__D_e", "EX_o2_e",
            reaction1_npoints=20, reaction2_npoints=20)
        self.assertEqual(data.growth_rates.shape, (20, 20))
        self.assertAlmostEqual(data.growth_rates.max(), 1.20898, places=4)
        self.assertAlmostEqual(abs(data.growth_rates[0, :]).max(), 0, places=4)
        if matplotlib is None:
            self.skip("can't test plots without matplotlib")
        data.plot()

# make a test suite to run all of the tests
loader = TestLoader()
suite = loader.loadTestsFromModule(sys.modules[__name__])


def test_all():
    TextTestRunner(verbosity=2).run(suite)

if __name__ == "__main__":
    test_all()



# set the warning format to be on a single line
import warnings as _warnings
from os.path import abspath as _abspath, dirname as _dirname
from os import name as _name

# set the warning format to be prettier and fit on one line
_cobra_path = _dirname(_abspath(__file__))
if _name == "posix":
    _warning_base = "%s:%s \x1b[1;31m%s\x1b[0m: %s\n"  # colors
else:
    _warning_base = "%s:%s %s: %s\n"


def _warn_format(message, category, filename, lineno, file=None, line=None):
    shortname = filename.replace(_cobra_path, "cobra", 1)
    return _warning_base % (shortname, lineno, category.__name__, message)
_warnings.formatwarning = _warn_format

from .version import get_version
__version__ = get_version()
from .core import Object, Metabolite, Gene, Reaction, Model, \
    DictList, Species
from . import io, flux_analysis, design

try:
    from .core import ArrayBasedModel
except ImportError:
    None

del get_version



"""Base class for sparse matrix formats using compressed storage
"""

__all__ = []

from warnings import warn

import numpy as np

from base import spmatrix, isspmatrix, SparseEfficiencyWarning
from data import _data_matrix
import sparsetools
from sputils import upcast, to_native, isdense, isshape, getdtype, \
        isscalarlike, isintlike


class _cs_matrix(_data_matrix):
    """base matrix class for compressed row and column oriented matrices"""

    def __init__(self, arg1, shape=None, dtype=None, copy=False, dims=None, nzmax=None):
        _data_matrix.__init__(self)

        if dims is not None:
            warn("dims= is deprecated, use shape= instead", DeprecationWarning)
            shape=dims

        if nzmax is not None:
            warn("nzmax= is deprecated", DeprecationWarning)


        if isspmatrix(arg1):
            if arg1.format == self.format and copy:
                arg1 = arg1.copy()
            else:
                arg1 = arg1.asformat(self.format)
            self._set_self( arg1 )

        elif isinstance(arg1, tuple):
            if isshape(arg1):
                # It's a tuple of matrix dimensions (M, N)
                # create empty matrix
                self.shape = arg1   #spmatrix checks for errors here
                M, N = self.shape
                self.data    = np.zeros(0, getdtype(dtype, default=float))
                self.indices = np.zeros(0, np.intc)
                self.indptr  = np.zeros(self._swap((M,N))[0] + 1, dtype=np.intc)
            else:
                if len(arg1) == 2:
                    # (data, ij) format
                    from coo import coo_matrix
                    other = self.__class__( coo_matrix(arg1, shape=shape) )
                    self._set_self( other )
                elif len(arg1) == 3:
                    # (data, indices, indptr) format
                    (data, indices, indptr) = arg1
                    self.indices = np.array(indices, copy=copy)
                    self.indptr  = np.array(indptr, copy=copy)
                    self.data    = np.array(data, copy=copy, dtype=getdtype(dtype, data))
                else:
                    raise ValueError, "unrecognized %s_matrix constructor usage" %\
                            self.format

        else:
            #must be dense
            try:
                arg1 = np.asarray(arg1)
            except:
                raise ValueError, "unrecognized %s_matrix constructor usage" % \
                        self.format
            from coo import coo_matrix
            self._set_self( self.__class__(coo_matrix(arg1, dtype=dtype)) )

        # Read matrix dimensions given, if any
        if shape is not None:
            self.shape = shape   # spmatrix will check for errors
        else:
            if self.shape is None:
                # shape not already set, try to infer dimensions
                try:
                    major_dim = len(self.indptr) - 1
                    minor_dim = self.indices.max() + 1
                except:
                    raise ValueError,'unable to infer matrix dimensions'
                else:
                    self.shape = self._swap((major_dim,minor_dim))

        if dtype is not None:
            self.data = self.data.astype(dtype)

        self.check_format(full_check=False)

    def getnnz(self):
        return self.indptr[-1]
    nnz = property(fget=getnnz)


    def _set_self(self, other, copy=False):
        """take the member variables of other and assign them to self"""

        if copy:
            other = other.copy()

        self.data    = other.data
        self.indices = other.indices
        self.indptr  = other.indptr
        self.shape   = other.shape

    def check_format(self, full_check=True):
        """check whether the matrix format is valid

        Parameters
        ==========

            - full_check : {bool}
                - True  - rigorous check, O(N) operations : default
                - False - basic check, O(1) operations

        """
        #use _swap to determine proper bounds
        major_name,minor_name = self._swap(('row','column'))
        major_dim,minor_dim = self._swap(self.shape)

        # index arrays should have integer data types
        if self.indptr.dtype.kind != 'i':
            warn("indptr array has non-integer dtype (%s)" \
                    % self.indptr.dtype.name )
        if self.indices.dtype.kind != 'i':
            warn("indices array has non-integer dtype (%s)" \
                    % self.indices.dtype.name )

        # only support 32-bit ints for now
        self.indptr  = np.asarray(self.indptr,  dtype=np.intc)
        self.indices = np.asarray(self.indices, dtype=np.intc)
        self.data    = to_native(self.data)

        # check array shapes
        if np.rank(self.data) != 1 or np.rank(self.indices) != 1 or np.rank(self.indptr) != 1:
            raise ValueError('data, indices, and indptr should be rank 1')

        # check index pointer
        if (len(self.indptr) != major_dim + 1 ):
            raise ValueError, \
                "index pointer size (%d) should be (%d)" % \
                 (len(self.indptr), major_dim + 1)
        if (self.indptr[0] != 0):
            raise ValueError,"index pointer should start with 0"

        # check index and data arrays
        if (len(self.indices) != len(self.data)):
            raise ValueError,"indices and data should have the same size"
        if (self.indptr[-1] > len(self.indices)):
            raise ValueError, \
                  "Last value of index pointer should be less than "\
                  "the size of index and data arrays"

        self.prune()

        if full_check:
            #check format validity (more expensive)
            if self.nnz > 0:
                if self.indices.max() >= minor_dim:
                    raise ValueError, "%s index values must be < %d" % \
                            (minor_name,minor_dim)
                if self.indices.min() < 0:
                    raise ValueError, "%s index values must be >= 0" % \
                            minor_name
                if np.diff(self.indptr).min() < 0:
                    raise ValueError,'index pointer values must form a " \
                                        "non-decreasing sequence'

        #if not self.has_sorted_indices():
        #    warn('Indices were not in sorted order.  Sorting indices.')
        #    self.sort_indices()
        #    assert(self.has_sorted_indices())
        #TODO check for duplicates?


    def __add__(self,other):
        # First check if argument is a scalar
        if isscalarlike(other):
            # Now we would add this scalar to every element.
            raise NotImplementedError, 'adding a scalar to a CSC or CSR ' \
                  'matrix is not supported'
        elif isspmatrix(other):
            if (other.shape != self.shape):
                raise ValueError, "inconsistent shapes"

            return self._binopt(other,'_plus_')
        elif isdense(other):
            # Convert this matrix to a dense matrix and add them
            return self.todense() + other
        else:
            raise NotImplementedError

    def __radd__(self,other):
        return self.__add__(other)

    def __sub__(self,other):
        # First check if argument is a scalar
        if isscalarlike(other):
            # Now we would add this scalar to every element.
            raise NotImplementedError, 'adding a scalar to a sparse ' \
                  'matrix is not supported'
        elif isspmatrix(other):
            if (other.shape != self.shape):
                raise ValueError, "inconsistent shapes"

            return self._binopt(other,'_minus_')
        elif isdense(other):
            # Convert this matrix to a dense matrix and subtract them
            return self.todense() - other
        else:
            raise NotImplementedError

    def __rsub__(self,other):  # other - self
        #note: this can't be replaced by other + (-self) for unsigned types
        if isscalarlike(other):
            # Now we would add this scalar to every element.
            raise NotImplementedError, 'adding a scalar to a sparse ' \
                  'matrix is not supported'
        elif isdense(other):
            # Convert this matrix to a dense matrix and subtract them
            return other - self.todense()
        else:
            raise NotImplementedError


    def __truediv__(self,other):
        if isscalarlike(other):
            return self * (1./other)

        elif isspmatrix(other):
            if other.shape != self.shape:
                raise ValueError('inconsistent shapes')

            return self._binopt(other,'_eldiv_')

        else:
            raise NotImplementedError


    def multiply(self, other):
        """Point-wise multiplication by another matrix
        """
        if other.shape != self.shape:
            raise ValueError('inconsistent shapes')

        if isdense(other):
            return np.multiply(self.todense(),other)
        else:
            other = self.__class__(other)
            return self._binopt(other,'_elmul_')


    ###########################
    # Multiplication handlers #
    ###########################

    def _mul_vector(self, other):
        M,N = self.shape

        #output array
        result = np.zeros( self.shape[0], dtype=upcast(self.dtype,other.dtype) )

        # csr_matvec or csc_matvec
        fn = getattr(sparsetools,self.format + '_matvec')
        fn(M, N, self.indptr, self.indices, self.data, other, result)

        return result


    def _mul_multivector(self, other):
        M,N = self.shape
        n_vecs = other.shape[1] #number of column vectors

        result = np.zeros( (M,n_vecs), dtype=upcast(self.dtype,other.dtype) )

        # csr_matvecs or csc_matvecs
        fn = getattr(sparsetools,self.format + '_matvecs')
        fn(M, N, n_vecs, self.indptr, self.indices, self.data, other.ravel(), result.ravel())

        return result


    def _mul_sparse_matrix(self, other):
        M, K1 = self.shape
        K2, N = other.shape

        major_axis = self._swap((M,N))[0]
        indptr = np.empty(major_axis + 1, dtype=np.intc)

        other = self.__class__(other) #convert to this format
        fn = getattr(sparsetools, self.format + '_matmat_pass1')
        fn( M, N, self.indptr, self.indices, \
                  other.indptr, other.indices, \
                  indptr)

        nnz = indptr[-1]
        indices = np.empty(nnz, dtype=np.intc)
        data    = np.empty(nnz, dtype=upcast(self.dtype,other.dtype))

        fn = getattr(sparsetools, self.format + '_matmat_pass2')
        fn( M, N, self.indptr, self.indices, self.data, \
                  other.indptr, other.indices, other.data, \
                  indptr, indices, data)

        return self.__class__((data,indices,indptr),shape=(M,N))


    @np.deprecate
    def getdata(self, ind):
        return self.data[ind]

    def diagonal(self):
        """Returns the main diagonal of the matrix
        """
        #TODO support k-th diagonal
        fn = getattr(sparsetools, self.format + "_diagonal")
        y = np.empty( min(self.shape), dtype=upcast(self.dtype) )
        fn(self.shape[0], self.shape[1], self.indptr, self.indices, self.data, y)
        return y

    def sum(self, axis=None):
        """Sum the matrix over the given axis.  If the axis is None, sum
        over both rows and columns, returning a scalar.
        """
        # The spmatrix base class already does axis=0 and axis=1 efficiently
        # so we only do the case axis=None here
        if axis is None:
            return self.data.sum()
        else:
            return spmatrix.sum(self,axis)
            raise ValueError, "axis out of bounds"

    #######################
    # Getting and Setting #
    #######################

    def __getitem__(self, key):
        if isinstance(key, tuple):
            row = key[0]
            col = key[1]

            #TODO implement CSR[ [1,2,3], X ] with sparse matmat
            #TODO make use of sorted indices

            if isintlike(row) and isintlike(col):
                return self._get_single_element(row,col)
            else:
                major,minor = self._swap((row,col))
                if isintlike(major) and isinstance(minor,slice):
                    minor_shape = self._swap(self.shape)[1]
                    start, stop, stride = minor.indices(minor_shape)
                    out_shape   = self._swap( (1, stop-start) )
                    return self._get_slice( major, start, stop, stride, out_shape)
                elif isinstance( row, slice) or isinstance(col, slice):
                    return self._get_submatrix( row, col )
                else:
                    raise NotImplementedError

        elif isintlike(key):
            return self[key, :]
        else:
            raise IndexError, "invalid index"


    def _get_single_element(self,row,col):
        M, N = self.shape
        if (row < 0):
            row += M
        if (col < 0):
            col += N
        if not (0<=row<M) or not (0<=col<N):
            raise IndexError("index out of bounds")

        major_index, minor_index = self._swap((row,col))

        start = self.indptr[major_index]
        end   = self.indptr[major_index+1]
        indxs = np.where(minor_index == self.indices[start:end])[0]

        num_matches = len(indxs)

        if num_matches == 0:
            # entry does not appear in the matrix
            return 0
        elif num_matches == 1:
            return self.data[start:end][indxs[0]]
        else:
            raise ValueError('nonzero entry (%d,%d) occurs more than once' % (row,col))

    def _get_slice(self, i, start, stop, stride, shape):
        """Returns a copy of the elements
            [i, start:stop:string] for row-oriented matrices
            [start:stop:string, i] for column-oriented matrices
        """
        if stride != 1:
            raise ValueError, "slicing with step != 1 not supported"
        if stop <= start:
            raise ValueError, "slice width must be >= 1"

        #TODO make [i,:] faster
        #TODO implement [i,x:y:z]

        indices = []

        for ind in xrange(self.indptr[i], self.indptr[i+1]):
            if self.indices[ind] >= start and self.indices[ind] < stop:
                indices.append(ind)

        index  = self.indices[indices] - start
        data   = self.data[indices]
        indptr = np.array([0, len(indices)])
        return self.__class__((data, index, indptr), shape=shape, \
                              dtype=self.dtype)

    def _get_submatrix( self, slice0, slice1 ):
        """Return a submatrix of this matrix (new matrix is created)."""

        slice0, slice1 = self._swap((slice0,slice1))
        shape0, shape1 = self._swap(self.shape)
        def _process_slice( sl, num ):
            if isinstance( sl, slice ):
                i0, i1 = sl.start, sl.stop
                if i0 is None:
                    i0 = 0
                elif i0 < 0:
                    i0 = num + i0

                if i1 is None:
                    i1 = num
                elif i1 < 0:
                    i1 = num + i1

                return i0, i1

            elif np.isscalar( sl ):
                if sl < 0:
                    sl += num

                return sl, sl + 1

            else:
                return sl[0], sl[1]

        def _in_bounds( i0, i1, num ):
            if not (0<=i0<num) or not (0<i1<=num) or not (i0<i1):
                raise IndexError,\
                      "index out of bounds: 0<=%d<%d, 0<=%d<%d, %d<%d" %\
                      (i0, num, i1, num, i0, i1)

        i0, i1 = _process_slice( slice0, shape0 )
        j0, j1 = _process_slice( slice1, shape1 )
        _in_bounds( i0, i1, shape0 )
        _in_bounds( j0, j1, shape1 )

        aux = sparsetools.get_csr_submatrix( shape0, shape1,
                                             self.indptr, self.indices,
                                             self.data,
                                             i0, i1, j0, j1 )

        data, indices, indptr = aux[2], aux[1], aux[0]
        shape = self._swap( (i1 - i0, j1 - j0) )

        return self.__class__( (data,indices,indptr), shape=shape )


    def __setitem__(self, key, val):
        if isinstance(key, tuple):
            row,col = key
            if not (isscalarlike(row) and isscalarlike(col)):
                raise NotImplementedError("Fancy indexing in assignment not "
                                          "supported for csr matrices.")
            M, N = self.shape
            if (row < 0):
                row += M
            if (col < 0):
                col += N
            if not (0<=row<M) or not (0<=col<N):
                raise IndexError, "index out of bounds"

            major_index, minor_index = self._swap((row,col))

            start = self.indptr[major_index]
            end   = self.indptr[major_index+1]
            indxs = np.where(minor_index == self.indices[start:end])[0]

            num_matches = len(indxs)


            if not np.isscalar(val):
                raise ValueError('setting an array element with a sequence')

            val = self.dtype.type(val)

            if num_matches == 0:
                #entry not already present
                warn('changing the sparsity structure of a %s_matrix is expensive. ' \
                        'lil_matrix is more efficient.' % self.format, \
                        SparseEfficiencyWarning)

                if self.has_sorted_indices:
                    # preserve sorted order
                    newindx = start + self.indices[start:end].searchsorted(minor_index)
                else:
                    newindx = start

                val         = np.array([val],         dtype=self.data.dtype)
                minor_index = np.array([minor_index], dtype=self.indices.dtype)

                self.data    = np.concatenate((self.data[:newindx],    val,         self.data[newindx:]))
                self.indices = np.concatenate((self.indices[:newindx], minor_index, self.indices[newindx:]))
                self.indptr  = self.indptr.copy()

                self.indptr[major_index+1:] += 1

            elif num_matches == 1:
                #entry appears exactly once
                self.data[start:end][indxs[0]] = val
            else:
                #entry appears more than once
                raise ValueError,'nonzero entry (%d,%d) occurs more than once' % (row,col)

            self.check_format(full_check=True)
        else:
            # We should allow slices here!
            raise IndexError, "invalid index"

    ######################
    # Conversion methods #
    ######################

    def todia(self):
        return self.tocoo(copy=False).todia()

    def todok(self):
        return self.tocoo(copy=False).todok()

    def tocoo(self,copy=True):
        """Return a COOrdinate representation of this matrix

        When copy=False the index and data arrays are not copied.
        """
        major_dim,minor_dim = self._swap(self.shape)

        data = self.data
        minor_indices = self.indices

        if copy:
            data = data.copy()
            minor_indices = minor_indices.copy()

        major_indices = np.empty(len(minor_indices), dtype=np.intc)

        sparsetools.expandptr(major_dim,self.indptr,major_indices)

        row,col = self._swap((major_indices,minor_indices))

        from coo import coo_matrix
        return coo_matrix((data,(row,col)), self.shape)

    def toarray(self):
        return self.tocoo(copy=False).toarray()

    ##############################################################
    # methods that examine or modify the internal data structure #
    ##############################################################

    def eliminate_zeros(self):
        """Remove zero entries from the matrix

        The is an *in place* operation
        """
        fn = sparsetools.csr_eliminate_zeros
        M,N = self._swap(self.shape)
        fn( M, N, self.indptr, self.indices, self.data)

        self.prune() #nnz may have changed

    def sum_duplicates(self):
        """Eliminate duplicate matrix entries by adding them together

        The is an *in place* operation
        """
        self.sort_indices()

        fn = sparsetools.csr_sum_duplicates
        M,N = self._swap(self.shape)
        fn( M, N, self.indptr, self.indices, self.data)

        self.prune() #nnz may have changed


    def __get_sorted(self):
        """Determine whether the matrix has sorted indices

        Returns
            - True: if the indices of the matrix are in sorted order
            - False: otherwise

        """

        #first check to see if result was cached
        if not hasattr(self,'__has_sorted_indices'):
            fn = sparsetools.csr_has_sorted_indices
            self.__has_sorted_indices = \
                    fn( len(self.indptr) - 1, self.indptr, self.indices)
        return self.__has_sorted_indices

    def __set_sorted(self, val):
        self.__has_sorted_indices = bool(val)

    has_sorted_indices = property(fget=__get_sorted, fset=__set_sorted)

    def sorted_indices(self):
        """Return a copy of this matrix with sorted indices
        """
        A = self.copy()
        A.sort_indices()
        return A

        # an alternative that has linear complexity is the following
        # although the previous option is typically faster
        #return self.toother().toother()

    def sort_indices(self):
        """Sort the indices of this matrix *in place*
        """

        if not self.has_sorted_indices:
            fn = sparsetools.csr_sort_indices
            fn( len(self.indptr) - 1, self.indptr, self.indices, self.data)
            self.has_sorted_indices = True

    #TODO remove after 0.7
    def ensure_sorted_indices(self, inplace=False):
        """Return a copy of this matrix where the column indices are sorted
        """
        warn('ensure_sorted_indices is deprecated, ' \
                'use sorted_indices() or sort_indices() instead', \
                DeprecationWarning)

        if inplace:
            self.sort_indices()
        else:
            return self.sorted_indices()

    def prune(self):
        """Remove empty space after all non-zero elements.
        """
        major_dim = self._swap(self.shape)[0]

        if len(self.indptr) != major_dim + 1:
            raise ValueError('index pointer has invalid length')
        if len(self.indices) < self.nnz:
            raise ValueError('indices array has fewer than nnz elements')
        if len(self.data) < self.nnz:
            raise ValueError('data array has fewer than nnz elements')

        self.data    = self.data[:self.nnz]
        self.indices = self.indices[:self.nnz]


    ###################
    # utility methods #
    ###################

    # needed by _data_matrix
    def _with_data(self,data,copy=True):
        """Returns a matrix with the same sparsity structure as self,
        but with different data.  By default the structure arrays
        (i.e. .indptr and .indices) are copied.
        """
        if copy:
            return self.__class__((data,self.indices.copy(),self.indptr.copy()), \
                                   shape=self.shape,dtype=data.dtype)
        else:
            return self.__class__((data,self.indices,self.indptr), \
                                   shape=self.shape,dtype=data.dtype)

    def _binopt(self, other, op):
        """apply the binary operation fn to two sparse matrices"""
        other = self.__class__(other)

        # e.g. csr_plus_csr, csr_minus_csr, etc.
        fn = getattr(sparsetools, self.format + op + self.format)

        maxnnz  = self.nnz + other.nnz
        indptr  = np.empty_like(self.indptr)
        indices = np.empty(maxnnz, dtype=np.intc)
        data    = np.empty(maxnnz, dtype=upcast(self.dtype,other.dtype))

        fn(self.shape[0], self.shape[1], \
                self.indptr,  self.indices,  self.data,
                other.indptr, other.indices, other.data,
                indptr, indices, data)

        actual_nnz = indptr[-1]
        indices = indices[:actual_nnz]
        data    = data[:actual_nnz]
        if actual_nnz < maxnnz / 2:
            #too much waste, trim arrays
            indices = indices.copy()
            data    = data.copy()

        A = self.__class__((data, indices, indptr), shape=self.shape)

        return A



#cobra.query.query.py
#Will serve as a location to house the growing number of
#simple query functions attached to cobra.Model

#NOTE: Many of the find functions are gone because Reactions,
#Metabolites, and Genes are now away of each other.

import re
#####
def print_reactions_involving_metabolite(cobra_model, the_metabolites):
    """Update to allow for multiple metabolite search

    cobra_model: A cobra.Model object

    the_metabolites: A list of cobra.Metabolites or metabolite ids that are in
    cobra_metabolites.

    #TODO: Move this to the Metabolite class

    """
    if hasattr(the_metabolites, 'id'):
        the_metabolites = [the_metabolites]
    elif not hasattr(the_metabolites, '__iter__'):
        the_metabolites = [the_metabolites]
    if not hasattr(the_metabolites[0], 'id'):
        the_metabolites = [cobra_model.metabolites[cobra_model.metabolites.index(x)]
                           for x in the_metabolites]
        
    for the_metabolite in the_metabolties:
        for the_reaction in the_metabolite._reaction:
            print the_reaction.reaction
 
         
def get_translation_reactions(cobra_model, genes_of_interest):
    """Find the translation elongation reactions for a set of genes
    in a cobra model.  Related to ME-model extensions

    cobra_model:  A cobra.Model object.

    genes_of_interest:  A list of genes from cobra_model.genes.
    
    """
    gene_translation_reactions = defaultdict(list)
    for the_reaction in cobra_model.reactions:
        if 'translation_elongation' in the_reaction:
            for the_gene in genes_of_interest:
                if the_gene in the_reaction:
                    gene_translation_reactions[the_gene].append(the_reaction)
                    continue
    return gene_translation_reactions


if __name__ == '__main__':
    from cPickle import load
    from time import time
    solver = 'glpk'
    test_directory = '../test/data/'
    with open(test_directory + 'salmonella.pickle') as in_file:
        cobra_model = load(in_file)

    #TODO: Add in tests for each function
    print 'Need to add in tests for %s'%repr(['print_reactions_involving_metabolite'])
                                              



# distutils: libraries=glpk
# cython: embedsignature=True

from glpk cimport *
from libc.stdlib cimport malloc, free
from cpython cimport bool
from cpython.version cimport PY_MAJOR_VERSION

from tempfile import NamedTemporaryFile as _NamedTemporaryFile  # for pickling
from os import unlink as _unlink
from warnings import warn as _warn

try:
    from sympy import Basic, Number
except:
    class Basic:
        pass
    Number = Basic

__glpk_version__ = str(glp_version())
_SUPPORTS_MILP = True
solver_name = "cglpk"

__doc__ = """Bindings to GLPK

The GNU Linear Programming Kit (GLPK) is released under the GPL.
The source can be downloaded from http://www.gnu.org/software/glpk/

The GNU Multiple Precision Arithmetic Library (GMP) is released under the GPL.
The source can be downloaded from https://gmplib.org/

"""

cdef dict ERROR_CODES = {
    GLP_EBADB: "GLP_EBADB",
    GLP_ESING: "GLP_ESING",
    GLP_ECOND: "GLP_ECOND",
    GLP_EBOUND: "GLP_EBOUND",
    GLP_EFAIL: "GLP_EFAIL",
    GLP_EOBJLL: "GLP_EOBJLL",
    GLP_EOBJUL: "GLP_EOBJUL",
    GLP_EITLIM: "GLP_EITLIM",
    GLP_ETMLIM: "GLP_ETMLIM",
    GLP_ENOPFS: "GLP_ENOPFS",
    GLP_ENODFS: "GLP_ENODFS",
    GLP_EROOT: "GLP_EROOT",
    GLP_ESTOP: "GLP_ESTOP",
    GLP_EMIPGAP: "GLP_EMIPGAP",
    GLP_ENOFEAS: "GLP_ENOFEAS",
    GLP_ENOCVG: "GLP_ENOCVG",
    GLP_EINSTAB: "GLP_EINSTAB",
    GLP_EDATA: "GLP_EDATA",
    GLP_ERANGE: "GLP_ERANGE"
}

cdef dict ERROR_MESSAGES = {
    GLP_EBADB: "invalid basis",
    GLP_ESING: "singular matrix",
    GLP_ECOND: "ill-conditioned matrix",
    GLP_EBOUND: "invalid bounds",
    GLP_EFAIL: "solver failed",
    GLP_EOBJLL: "objective lower limit reached",
    GLP_EOBJUL: "objective upper limit reached",
    GLP_EITLIM: "iteration limit exceeded",
    GLP_ETMLIM: "time limit exceeded",
    GLP_ENOPFS: "no primal feasible solution",
    GLP_ENODFS: "no dual feasible solution",
    GLP_EROOT: "root LP optimum not provided",
    GLP_ESTOP: "search terminated by application",
    GLP_EMIPGAP: "relative mip gap tolerance reached",
    GLP_ENOFEAS: "no primal/dual feasible solution",
    GLP_ENOCVG: "no convergence",
    GLP_EINSTAB: "numerical instability",
    GLP_EDATA: "invalid data",
    GLP_ERANGE: "result out of range"
}

cdef dict METHODS = {
    "auto": GLP_DUALP,
    "primal": GLP_PRIMAL,
    "dual": GLP_DUAL
}

cdef dict PRICINGS = {
    "std": GLP_PT_STD,
    "pse": GLP_PT_PSE
}

cdef dict RATIOS = {
    "std": GLP_RT_STD,
    "har": GLP_RT_HAR
}

cdef dict SCALINGS = {
    "auto": GLP_SF_AUTO,
    "gm": GLP_SF_GM,
    "eq": GLP_SF_EQ,
    "2n": GLP_SF_2N,
    "skip": GLP_SF_SKIP
}

cdef int downcast_pos_size(Py_ssize_t size):
    if size > INT_MAX:
        raise ValueError("Integer overflow %d > %d" % (size, INT_MAX))
    else:
        return <int> size


cdef check_error(int result):
    if result == 0:
        return
    if result not in ERROR_CODES:
        raise RuntimeError("glp_simplex failed with unknown error code 0x%x" %
                           result)
    raise RuntimeError("glp_simplex failed with error code %s: %s" %
                       (ERROR_CODES[result], ERROR_MESSAGES[result]))



cdef int hook(void *info, const char *s):
    """function to redirect sdout to python stdout"""
    print(s)
    return 1


cdef int silent_hook(void *info, const char *s):
    """function to print nothing but trick GLPK into thinking we did"""
    return 1

cdef double _to_double(value):
    if isinstance(value, Basic) and not isinstance(value, Number):
        return 0.
    else:
        return <double>value


cdef class GLP:
    cdef glp_prob *glp
    cdef glp_smcp parameters
    cdef glp_iocp integer_parameters
    cdef public bool exact

    # cython related allocation/dellocation functions
    def __cinit__(self):
        self.glp = glp_create_prob()
        # initialize parameters
        glp_set_obj_dir(self.glp, GLP_MAX)  # default is maximize
        glp_init_smcp(&self.parameters)
        glp_init_iocp(&self.integer_parameters)
        self.exact = False
        glp_term_hook(hook, NULL)
        self.parameters.msg_lev = GLP_MSG_OFF
        self.integer_parameters.tol_int = 1e-9

    def __dealloc__(self):
        glp_delete_prob(self.glp)

    def __init__(self, cobra_model=None):
        cdef int bound_type, index, n, n_values
        cdef glp_prob *glp
        cdef int *c_rows
        cdef int *c_cols
        cdef double *c_values
        cdef double b

        if cobra_model is None:
            return
        glp = self.glp
        glp_add_rows(glp, downcast_pos_size(len(cobra_model.metabolites)))
        glp_add_cols(glp, downcast_pos_size(len(cobra_model.reactions)))

        metabolite_id_to_index = {r.id: index for index, r
                                  in enumerate(cobra_model.metabolites, 1)}

        linear_constraint_rows = []
        linear_constraint_cols = []
        linear_constraint_values = []

        # set metabolite/consraint bounds
        for index, metabolite in enumerate(cobra_model.metabolites, 1):
            b = _to_double(metabolite._bound)
            c = metabolite._constraint_sense
            if c == 'E':
                bound_type = GLP_FX  # Set metabolite to steady state levels
            elif c == 'L':
                bound_type = GLP_UP  # x < 2 <==> x has an upper bound of 2
            elif c == 'G':
                bound_type = GLP_LO  # x > 2 <==> x has a lower bound of 2
            else:
                raise ValueError("unsupported bound type: %s" % c)
            glp_set_row_bnds(glp, index, bound_type, b, b)
        
        # set reaction/varaiable bounds
        for index, reaction in enumerate(cobra_model.reactions, 1):
            if reaction.variable_kind == "integer":
                if reaction.lower_bound == 0 and reaction.upper_bound == 1:
                    glp_set_col_kind(self.glp, index, GLP_BV)  # binary
                else:
                    glp_set_col_kind(self.glp, index, GLP_IV)
            if reaction.lower_bound == reaction.upper_bound:
                bound_type = GLP_FX
            else:
                bound_type = GLP_DB
            glp_set_col_bnds(glp, index, bound_type,
                             _to_double(reaction.lower_bound),
                             _to_double(reaction.upper_bound))
            glp_set_obj_coef(glp, index,
                             _to_double(reaction.objective_coefficient))

            for metabolite, coefficient in reaction._metabolites.iteritems():
                linear_constraint_rows.append(
                    metabolite_id_to_index[metabolite.id])
                linear_constraint_cols.append(index)
                linear_constraint_values.append(coefficient)
        
        # set constraint marix
        # first copy the python lists to c arrays
        n_values = downcast_pos_size(len(linear_constraint_rows))
        c_cols = <int *> malloc((n_values + 1) * sizeof(int))
        c_rows = <int *> malloc((n_values + 1) * sizeof(int))
        c_values = <double *> malloc((n_values + 1) * sizeof(double))
        if c_rows is NULL or c_rows is NULL or c_values is NULL:
            raise MemoryError()
        for index in range(n_values):
            c_rows[index + 1] = linear_constraint_rows[index]
            c_cols[index + 1] = linear_constraint_cols[index]
            c_values[index + 1] = _to_double(linear_constraint_values[index])
        # actually set the values
        glp_load_matrix(glp, n_values, c_rows, c_cols, c_values)
        # free the c arrays
        free(c_rows)
        free(c_cols)
        free(c_values)

    # problem creation and modification
    @classmethod
    def create_problem(cls, cobra_model, objective_sense="maximize"):
        problem = cls(cobra_model)
        problem.set_objective_sense(objective_sense)
        return problem

    cpdef change_variable_bounds(self, int index, double lower_bound,
                                 double upper_bound):
        cdef int bound_type = GLP_FX if lower_bound == upper_bound else GLP_DB
        assert index >= 0
        glp_set_col_bnds(self.glp, index + 1, bound_type, lower_bound, upper_bound)

    def change_coefficient(self, int met_index, int rxn_index, double value):
        cdef int col_length, i
        cdef int *indexes
        cdef double *values
        # glpk uses 1 indexing
        met_index += 1
        rxn_index += 1
        # we first have to get the old column
        col_length = glp_get_mat_col(self.glp, rxn_index, NULL, NULL)
        indexes = <int *> malloc((col_length + 2) * sizeof(int))
        values = <double *> malloc((col_length + 2) * sizeof(double))
        if indexes == NULL or values == NULL:
            raise MemoryError()
        glp_get_mat_col(self.glp, rxn_index, indexes, values)
        # search for duplicate
        for i in range(col_length):
            # if a duplicate exists replace that value and exit
            if indexes[i + 1] == met_index:
                values[i + 1] = value
                glp_set_mat_col(self.glp, rxn_index, col_length, indexes, values)
                return
        # need to add a new entry
        indexes[col_length + 1] = met_index
        values[col_length + 1] = value
        glp_set_mat_col(self.glp, rxn_index, col_length + 1, indexes, values)
        free(indexes)
        free(values)

    def solve_problem(self, **solver_parameters):
        cdef int result
        cdef int time_limit
        cdef glp_prob *glp = self.glp

        for key, value in solver_parameters.items():
            self.set_parameter(key, value)

        # suspend the gil to allow multithreading
        # multithreading must occur with DIFFERENT glp objects
        # calling solve_problem on the same object from 2 different
        # threads at the same time will probably cause problems
        # because glpk itself is not thread safe

        #with nogil:  # we can use this if glpk ever gets thread-safe malloc
        # Try to solve the problem with the existing basis, but with
        # a time limit in case it gets stuck.
        time_limit = self.parameters.tm_lim  # save time limit
        self.parameters.tm_lim = min(500, time_limit)
        fast_status = glp_simplex(glp, &self.parameters)
        self.parameters.tm_lim = time_limit
        
        if fast_status != 0:
            glp_adv_basis(glp, 0)
            check_error(glp_simplex(glp, &self.parameters))
        self.parameters.tm_lim = time_limit
        if self.exact:
            # sigh... it looks like the exact routine doesn't fully respect
            # the verbosity parameter
            if self.parameters.msg_lev == GLP_MSG_OFF:
                glp_term_hook(silent_hook, NULL)
            try:
                check_error(glp_exact(glp, &self.parameters))
            finally:
                glp_term_hook(hook, NULL)

        if self.is_mip():
            self.integer_parameters.tm_lim = self.parameters.tm_lim
            self.integer_parameters.msg_lev = self.parameters.msg_lev
            #with nogil:
            check_error(glp_intopt(glp, &self.integer_parameters))
        return self.get_status()

    @classmethod
    def solve(cls, cobra_model, **kwargs):
        problem = cls.create_problem(cobra_model)
        problem.solve_problem(**kwargs)
        solution = problem.format_solution(cobra_model)
        return solution

    def get_status(self):
        cdef int result = glp_mip_status(self.glp) if self.is_mip() \
            else glp_get_status(self.glp)
        if result == GLP_OPT:
            return "optimal"
        if result == GLP_FEAS:
            return glp_get_status(self.glp)
        if result == GLP_UNDEF:
            return "undefined"
        if result == GLP_UNBND:
            return "unbounded"
        if result == GLP_NOFEAS:
            return "infeasible"
        return "failed"

    cpdef set_objective_sense(self, objective_sense):
        objective_sense = objective_sense.lower()
        if objective_sense == "maximize":
            glp_set_obj_dir(self.glp, GLP_MAX)
        elif objective_sense == "minimize":
            glp_set_obj_dir(self.glp, GLP_MIN)
        else:
            raise ValueError("%s is not a valid objective sense" % objective_sense)

    cpdef set_parameter(self, parameter_name, value):
        """set a solver parameter"""
        if parameter_name == "objective_sense":
            self.set_objective_sense(value)
        elif parameter_name in {"time_limit", "tm_lim"}:
            self.parameters.tm_lim = int(1000 * value)
            # Setting a value less than 0.001 would cause us to not
            # set a time limit at all. It's better to set a time limit
            # of 1 ms in this case.
            #if value > 0 and self.parameters.tm_lim == 0:
            #    self.parameters.tm_lim = 1
        elif parameter_name == "tolerance_feasibility":
            self.parameters.tol_bnd = float(value)
            self.parameters.tol_dj = float(value)
        elif parameter_name == "tol_bnd":
            self.parameters.tol_bnd = float(value)
        elif parameter_name == "tol_dj":
            self.parameters.tol_dj = float(value)
        elif parameter_name in {"tolerance_markowitz", "tol_piv"}:
            self.parameters.tol_piv = float(value)
        elif parameter_name in {"tolerance_integer", "tol_int"}:
            self.integer_parameters.tol_int = float(value)
        elif parameter_name in {"mip_gap", "MIP_gap"}:
            self.integer_parameters.mip_gap = float(value)
        elif parameter_name == "verbose":
            if not value:  # suppress all output
                self.parameters.msg_lev = GLP_MSG_OFF
                return
            if value == "err":
                self.parameters.msg_lev = GLP_MSG_ERR
            elif value is True or value == "all":
                self.parameters.msg_lev = GLP_MSG_ALL
            elif value == "normal":
                self.parameters.msg_lev = GLP_MSG_ON
        elif parameter_name == "iteration_limit":
            self.parameters.it_lim = value
        elif parameter_name == "lp_method":
            self.parameters.meth = METHODS[value]
        elif parameter_name == "exact":
            self.exact = value
        elif parameter_name == "threads":
            _warn("multiple threads not supported")
        elif parameter_name == "MIP_gap_abs":
            _warn("setting aboslute mip gap not supported")
        elif parameter_name == "presolve":
            self.parameters.presolve = GLP_ON if value else GLP_OFF
        elif parameter_name == "pricing":
            self.parameters.pricing = PRICINGS[value]
        elif parameter_name == "r_test":
            self.parameters.r_test = RATIOS[value]
        elif parameter_name == "scale":
            if value:
                if isinstance(value, int):
                    glp_scale_prob(self.glp, value)
                else:
                    glp_scale_prob(self.glp, SCALINGS[value])
            else:
                glp_unscale_prob(self.glp)
        elif parameter_name == "quadratic_component":
            if value is not None:
                raise ValueError("quadratic component must be None for glpk")
        else:
            raise ValueError("unknown parameter " + str(parameter_name))

    cpdef get_objective_value(self):
        if self.is_mip():
            return glp_mip_obj_val(self.glp)
        return glp_get_obj_val(self.glp)

    cpdef change_variable_objective(self, int index, double value):
        assert index >= 0
        glp_set_obj_coef(self.glp, index + 1, value)

    cpdef is_mip(self):
        return glp_get_num_int(self.glp) > 0

    def format_solution(self, cobra_model):
        cdef int i, m, n
        cdef glp_prob *glp = self.glp
        Solution = cobra_model.solution.__class__
        status = self.get_status()
        if status != "optimal":  # todo handle other possible
            return Solution(None, status=status)
        solution = Solution(self.get_objective_value(), status=status)
        m = glp_get_num_rows(glp)
        n = glp_get_num_cols(glp)
        x = [0] * n
        if self.is_mip():
            for i in range(1, n + 1):
                    x[i - 1] = glp_mip_col_val(glp, i)
            solution.x_dict = {rxn.id: x[i] for i, rxn
                               in enumerate(cobra_model.reactions)}
            solution.x = x
        else:
            for i in range(1, n + 1):
                x[i - 1] = glp_get_col_prim(glp, i)
            solution.x_dict = {rxn.id: x[i] for i, rxn
                               in enumerate(cobra_model.reactions)}
            solution.x = x
            y = [0] * m
            for i in range(1, m + 1):
                y[i - 1] = glp_get_row_dual(glp, i)
            solution.y_dict = {met.id: y[i] for i, met
                               in enumerate(cobra_model.metabolites)}
            solution.y = y
        return solution

    # make serializable and copyable
    def __getstate__(self):
        cdef int result
        cdef char *name
        tempfile = _NamedTemporaryFile(mode="r", delete=False)
        name = tempfile.name
        tempfile.close()
        result = glp_write_prob(self.glp, 0, name)
        assert result == 0
        with open(name, "r") as infile:
            state = infile.read()
        _unlink(name)
        return state

    def __reduce__(self):
        return (GLP, (), self.__getstate__())

    def __setstate__(self, state):
        cdef int result
        cdef char *name = NULL
        with _NamedTemporaryFile(mode="w", delete=False) as tempfile:
            name = tempfile.name
            tempfile.write(state)
        result = glp_read_prob(self.glp, 0, name)
        assert result == 0
        _unlink(name)

    def __copy__(self):
        other = GLP()
        glp_copy_prob(other.glp, self.glp, GLP_ON)
        other.parameters = self.parameters
        other.integer_parameters = self.integer_parameters
        other.exact = self.exact
        return other

    def write(self, filename):
        """Write the problem to a file

        The format will be determined by the file extension.
            .lp: LP file
            .mps: MPS file
        """
        if PY_MAJOR_VERSION == 2:
            b_name = bytes(filename)
        elif PY_MAJOR_VERSION == 3:
            b_name = bytes(filename, "latin-1")
        else:
            raise RuntimeError("Unknown python version")
        cdef char *c_name = <bytes> b_name
        cdef int res
        # no other way to silence this function
        glp_term_hook(silent_hook, NULL)
        try:
            if b_name.endswith(".lp"):
                res = glp_write_lp(self.glp, NULL, c_name)
            elif b_name.endswith(".mps"):
                res = glp_write_mps(self.glp, GLP_MPS_FILE, NULL, c_name)
            else:
                raise ValueError("Unknown file format for %s" % str(filename))
            if res != 0:
                raise IOError("failed to write LP to file %s" % str(filename))
        finally:
            glp_term_hook(hook, NULL)


# wrappers for all the functions at the module level
create_problem = GLP.create_problem
def set_objective_sense(lp, objective_sense="maximize"):
    return lp.set_objective_sense(lp, objective_sense=objective_sense)
cpdef change_variable_bounds(lp, int index,
                             double lower_bound, double upper_bound):
    return lp.change_variable_bounds(index, lower_bound, upper_bound)
cpdef change_variable_objective(lp, int index, double value):
    return lp.change_variable_objective(index, value)
cpdef change_coefficient(lp, int met_index, int rxn_index, double value):
    return lp.change_coefficient(met_index, rxn_index, value)
cpdef set_parameter(lp, parameter_name, value):
    return lp.set_parameter(parameter_name, value)
def solve_problem(lp, **kwargs):
    return lp.solve_problem(**kwargs)
cpdef get_status(lp):
    return lp.get_status()
cpdef get_objective_value(lp):
    return lp.get_objective_value()
cpdef format_solution(lp, cobra_model):
    return lp.format_solution(cobra_model)
solve = GLP.solve



#cobra.flux_analysis.reaction.py
#functions for analyzing / creating objective functions
from ..core.Reaction import Reaction
from six import iteritems

def assess(model, reaction, flux_coefficient_cutoff=0.001):
    """Assesses the capacity of the model to produce the precursors for the reaction
    and absorb the production of the reaction while the reaction is operating at, or
    above, the specified cutoff.

    model: A :class:`~cobra.core.Model` object

    reaction: A :class:`~cobra.core.Reaction` object


    flux_coefficient_cutoff:  Float.  The minimum flux that reaction must carry to
    be considered active.

    returns: True if the model can produce the precursors and absorb the products
    for the reaction operating at, or above, flux_coefficient_cutoff.  Otherwise,
    a dictionary of {'precursor': Status, 'product': Status}.  Where Status is the
    results from assess_precursors and assess_products,
    respectively.

    """
    reaction = model.reactions.get_by_id(reaction.id)
    model.optimize(new_objective={reaction: 1})
    if model.solution.f >= flux_coefficient_cutoff:
        return(True)
    else:
        results = {}
        results['precursors'] = assess_precursors(model, reaction, flux_coefficient_cutoff)
        results['products'] = assess_products(model, reaction, flux_coefficient_cutoff)
        return(results)

def assess_precursors(model, reaction,  flux_coefficient_cutoff=0.001):
    """Assesses the ability of the model to provide sufficient precursors for
    a reaction operating at, or beyond, the specified cutoff.

    model: A :class:`~cobra.core.Model` object

    reaction: A :class:`~cobra.core.Reaction` object


    flux_coefficient_cutoff:  Float.  The minimum flux that reaction must carry to
    be considered active.

    returns: True if  the precursors can be simultaneously produced at the specified cutoff.  False, if the model has
    the capacity to produce each individual precursor at the specified threshold  but not all precursors at the
    required level simultaneously.   Otherwise a dictionary of the required and the produced
    fluxes for each reactant that is not produced in sufficient quantities.
    
    """
    model = model.copy()
    reaction = model.reactions.get_by_id(reaction.id)
    model.optimize(new_objective={reaction: 1})
    if model.solution.f >= flux_coefficient_cutoff:
        return(True)
    #
    simulation_results = {}
    #build the sink reactions and add all at once
    sink_reactions = {}
    for the_component in reaction.get_reactants():
        #add in a sink reaction for each component
        sink_reaction = Reaction('test_sink_%s'%the_component.id)
        #then simulate production ability
        #then check it can exceed objective cutoff * component stoichiometric
        #coefficient.
        coefficient = reaction.get_coefficient(the_component) 
        sink_reaction.add_metabolites({the_component: coefficient})
        sink_reaction.upper_bound = 1000
        sink_reactions[sink_reaction] = (the_component, coefficient)
    #First assess whether all precursors can pbe produced simultaneously
    super_sink = Reaction("super_sink")
    for reaction in sink_reactions:
        super_sink += reaction
    super_sink.id = 'super_sink'
    model.add_reactions(sink_reactions.keys() + [super_sink])
    model.optimize(new_objective=super_sink)
    if flux_coefficient_cutoff <= model.solution.f:
        return(True)

    #Otherwise assess the ability of the model to produce each precursor individually.
    #Now assess the ability of the model to produce each reactant for a reaction
    for sink_reaction, (component, coefficient) in iteritems(sink_reactions):
        model.optimize(new_objective=sink_reaction) #Calculate the maximum amount of the
        #metabolite that can be produced.
        if flux_coefficient_cutoff > model.solution.f:
            #Scale the results to a single unit
            simulation_results.update({component:{'required':flux_coefficient_cutoff/abs(coefficient),
                                                  'produced':model.solution.f/abs(coefficient)}})
    if len(simulation_results) == 0:
        simulation_results = False
    return(simulation_results)

def assess_products(model, reaction, flux_coefficient_cutoff=0.001):
    """Assesses whether the model has the capacity to absorb the products of a reaction
    at a given flux rate.  Useful for identifying which components might be blocking
    a reaction from achieving a specific flux rate.

    model: A :class:`~cobra.core.Model` object

    reaction: A :class:`~cobra.core.Reaction` object

    flux_coefficient_cutoff:  Float.  The minimum flux that reaction must carry to
    be considered active.

    returns: True if the model has the capacity to absorb all the reaction products being
    simultaneously given the specified cutoff.   False, if the model has
    the capacity to absorb each individual product but not all products at the
    required level simultaneously.   Otherwise a dictionary of the required and the
    capacity fluxes for each product that is not absorbed in sufficient quantities.

    
    """
    model = model.copy()
    reaction = model.reactions.get_by_id(reaction.id)
    model.optimize(new_objective={reaction: 1})
    if model.solution.f >= flux_coefficient_cutoff:
        return(True)
    #
    simulation_results = {}
    #build the sink reactions and add all at once
    source_reactions = {}
    for the_component in reaction.get_products():
        #add in a sink reaction for each component
        source_reaction = Reaction('test_source_%s'%the_component.id)
        #then simulate production ability
        #then check it can exceed objective cutoff * component stoichiometric
        #coefficient.
        coefficient = reaction.get_coefficient(the_component) 
        source_reaction.add_metabolites({the_component: coefficient})
        source_reaction.upper_bound = 1000
        source_reactions[source_reaction] = (the_component, coefficient)
    #
    super_source = Reaction('super_source')
    for reaction in source_reactions:
            super_source += reaction
    super_source.id = 'super_source'
    model.add_reactions(source_reactions.keys() + [super_source])
    model.optimize(new_objective=super_source)
    if flux_coefficient_cutoff <= model.solution.f:
        return(True)

    #Now assess the ability of the model to produce each reactant for a reaction
    for source_reaction, (component, coefficient) in iteritems(source_reactions):
        model.optimize(new_objective=source_reaction) #Calculate the maximum amount of the
        #metabolite that can be produced.
        if flux_coefficient_cutoff > model.solution.f:
            #Scale the results to a single unit
            simulation_results.update({component:{'required':flux_coefficient_cutoff/abs(coefficient),
                                                  'capacity':model.solution.f/abs(coefficient)}})
    if len(simulation_results) == 0:
        simulation_results = False
    return(simulation_results)

    
    



from six import iteritems


class Object(object):
    """Defines common behavior of object in cobra.core"""

    def __init__(self, id=None, name=""):
        """
        id: None or a string

        """
        self.id = id
        self.name = name

        self.notes = {}
        self.annotation = {}

    def __getstate__(self):
        """To prevent excessive replication during deepcopy."""
        state = self.__dict__.copy()
        if '_model' in state:
            state['_model'] = None
        return state

    def __repr__(self):
        return "<%s %s at 0x%x>" % (self.__class__.__name__, self.id, id(self))

    def __str__(self):
        return str(self.id)



from warnings import warn
import re

from .Species import Species

# Numbers are not required because of the |(?=[A-Z])? block. See the
# discussion in https://github.com/opencobra/cobrapy/issues/128 for
# more details.
element_re = re.compile("([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)")


class Metabolite(Species):
    """Metabolite is a class for holding information regarding
    a metabolite in a cobra.Reaction object.

    """

    def __init__(self, id=None, formula=None,
                 name="", compartment=None):
        """
        id: str

        formula: str
            Chemical formula (i.e. H2O)

        name: str
            A human readable name.

        compartment: None or a dictionary indicating the cellular location
        of the metabolite.  Used when in a cobra.Reaction or Model
        object

        """
        Species.__init__(self, id, name)
        self.formula = formula
        # because in a Model a metabolite may participate in multiple Reactions
        self.compartment = compartment
        self.charge = None

        self._constraint_sense = 'E'
        self._bound = 0.

    @property
    def elements(self):
        tmp_formula = self.formula
        if tmp_formula is None:
            return {}
        # necessary for some old pickles which use the deprecated
        # Formula class
        tmp_formula = str(self.formula)
        # commonly occuring characters in incorrectly constructed formulas
        if "*" in tmp_formula:
            warn("invalid character '*' found in formula '%s'" % self.formula)
            tmp_formula = tmp_formula.replace("*", "")
        if "(" in tmp_formula or ")" in tmp_formula:
            warn("invalid formula (has parenthesis) in '%s'" % self.formula)
            return None
        composition = {}
        parsed = element_re.findall(tmp_formula)
        for (element, count) in parsed:
            if count == '':
                count = 1
            else:
                try:
                    count = float(count)
                    int_count = int(count)
                    if count == int_count:
                        count = int_count
                    else:
                        warn("%s is not an integer (in formula %s)" %
                             (count, self.formula))
                except ValueError:
                    warn("failed to parse %s (in formula %s)" %
                         (count, self.formula))
                    return None
            if element in composition:
                composition[element] += count
            else:
                composition[element] = count
        return composition

    @property
    def formula_weight(self):
        """Calculate the formula weight"""
        weight_dict = elements_and_molecular_weights
        try:
            return sum([count * elements_and_molecular_weights[element]
                        for element, count in self.elements.items()])
        except KeyError as e:
            warn("The element %s does not appear in the peridic table" % e)

    @property
    def y(self):
        """The shadow price for the metabolite in the most recent solution

        Shadow prices are computed from the dual values of the bounds in
        the solution.

        """
        try:
            return self._model.solution.y_dict[self.id]
        except Exception as e:
            if self._model is None:
                raise Exception("not part of a model")
            if not hasattr(self._model, "solution") or \
                    self._model.solution is None or \
                    self._model.solution.status == "NA":
                raise Exception("model has not been solved")
            if self._model.solution.status != "optimal":
                raise Exception("model solution was not optimal")
            raise e  # Not sure what the exact problem was

    def remove_from_model(self, method='subtractive', **kwargs):
        """Removes the association from self.model

        method: 'subtractive' or 'destructive'.
            If 'subtractive' then the metabolite is removed from all
            associated reactions.  If 'destructive' then all associated
            reactions are removed from the Model.

        """
        # why is model being taken in as a parameter? This plays
        # back to the question of allowing a Metabolite to be associated
        # with multiple Models
        if "model" in kwargs:
            warn("model argument deprecated")

        self._model.metabolites.remove(self)
        self._model = None
        if method.lower() == 'subtractive':
            for the_reaction in list(self._reaction):
                the_coefficient = the_reaction._metabolites[self]
                the_reaction.subtract_metabolites({self: the_coefficient})
        elif method.lower() == 'destructive':
            for x in self._reaction:
                x.remove_from_model()
        else:
            raise Exception(method + " is not 'subtractive' or 'destructive'")


elements_and_molecular_weights = {
    'H':   1.007940,
    'He':  4.002602,
    'Li':  6.941000,
    'Be':  9.012182,
    'B':   10.811000,
    'C':   12.010700,
    'N':   14.006700,
    'O':   15.999400,
    'F':   18.998403,
    'Ne':  20.179700,
    'Na':  22.989770,
    'Mg':  24.305000,
    'Al':  26.981538,
    'Si':  28.085500,
    'P':   30.973761,
    'S':   32.065000,
    'Cl':  35.453000,
    'Ar':  39.948000,
    'K':   39.098300,
    'Ca':  40.078000,
    'Sc':  44.955910,
    'Ti':  47.867000,
    'V':   50.941500,
    'Cr':  51.996100,
    'Mn':  54.938049,
    'Fe':  55.845000,
    'Co':  58.933200,
    'Ni':  58.693400,
    'Cu':  63.546000,
    'Zn':  65.409000,
    'Ga':  69.723000,
    'Ge':  72.640000,
    'As':  74.921600,
    'Se':  78.960000,
    'Br':  79.904000,
    'Kr':  83.798000,
    'Rb':  85.467800,
    'Sr':  87.620000,
    'Y':   88.905850,
    'Zr':  91.224000,
    'Nb':  92.906380,
    'Mo':  95.940000,
    'Tc':  98.000000,
    'Ru':  101.070000,
    'Rh':  102.905500,
    'Pd':  106.420000,
    'Ag':  107.868200,
    'Cd':  112.411000,
    'In':  114.818000,
    'Sn':  118.710000,
    'Sb':  121.760000,
    'Te':  127.600000,
    'I':   126.904470,
    'Xe':  131.293000,
    'Cs':  132.905450,
    'Ba':  137.327000,
    'La':  138.905500,
    'Ce':  140.116000,
    'Pr':  140.907650,
    'Nd':  144.240000,
    'Pm':  145.000000,
    'Sm':  150.360000,
    'Eu':  151.964000,
    'Gd':  157.250000,
    'Tb':  158.925340,
    'Dy':  162.500000,
    'Ho':  164.930320,
    'Er':  167.259000,
    'Tm':  168.934210,
    'Yb':  173.040000,
    'Lu':  174.967000,
    'Hf':  178.490000,
    'Ta':  180.947900,
    'W':   183.840000,
    'Re':  186.207000,
    'Os':  190.230000,
    'Ir':  192.217000,
    'Pt':  195.078000,
    'Au':  196.966550,
    'Hg':  200.590000,
    'Tl':  204.383300,
    'Pb':  207.200000,
    'Bi':  208.980380,
    'Po':  209.000000,
    'At':  210.000000,
    'Rn':  222.000000,
    'Fr':  223.000000,
    'Ra':  226.000000,
    'Ac':  227.000000,
    'Th':  232.038100,
    'Pa':  231.035880,
    'U':   238.028910,
    'Np':  237.000000,
    'Pu':  244.000000,
    'Am':  243.000000,
    'Cm':  247.000000,
    'Bk':  247.000000,
    'Cf':  251.000000,
    'Es':  252.000000,
    'Fm':  257.000000,
    'Md':  258.000000,
    'No':  259.000000,
    'Lr':  262.000000,
    'Rf':  261.000000,
    'Db':  262.000000,
    'Sg':  266.000000,
    'Bh':  264.000000,
    'Hs':  277.000000,
    'Mt':  268.000000,
    'Ds':  281.000000,
    'Rg':  272.000000,
    'Cn':  285.000000,
    'Uuq': 289.000000,
    'Uuh': 292.000000
}



from .DictList import DictList
from .Object import Object
from .Gene import Gene
from .Metabolite import Metabolite
from .Reaction import Reaction
from .Solution import Solution
from .Model import Model
from .Species import Species

try:
    import scipy
except:
    scipy = None

if scipy:
    from .ArrayBasedModel import ArrayBasedModel
else:
    from warnings import warn
    warn("ArrayBasedModel requires scipy")
    del warn
del scipy
